"""Shared utilities for doing decryption.
"""

from ploigos_step_runner.exceptions import StepRunnerException
from ploigos_step_runner.utils.io import TextIOSelectiveObfuscator
from ploigos_step_runner.utils.reflection import import_and_get_class
from ploigos_step_runner.config.config_value_decryptor import ConfigValueDecryptor

class DecryptionUtils:
    """Shared utilities for doing decryption.

    Any values that are decrypted are added to the given list of TextIOSelectiveObfuscator
    of strings to obfuscate.

    Attributes
    ----------
    __obfuscation_streams : list of TextIOSelectiveObfuscator
        TextIOSelectiveObfuscators to be sure that any decrypted values are obfuscated
        on those streams.
    __config_value_decryptors : list of ConfigValueDecryptor
        ConfigValueDecryptors that can be used to decrypt given ConfigValue.
    """

    __DEFAULT_DECRYPTORS_MODULE = 'ploigos_step_runner.config.decryptors'

    __obfuscation_streams = []
    __config_value_decryptors = []

    @staticmethod
    def register_obfuscation_stream(obfuscator_stream):
        """Add a TextIOSelectiveObfuscator to obfuscate any decrypted values on.

        Parameters
        ----------
        obfuscator_stream : TextIOSelectiveObfuscator
            TextIOSelectiveObfuscator to be sure that any decrypted values are obfuscated on.

        Raises
        ------
        AssertionError
            If given obfuscator_stream is not a type of TextIOSelectiveObfuscator
        """
        assert isinstance(obfuscator_stream, TextIOSelectiveObfuscator)
        DecryptionUtils.__obfuscation_streams.append(obfuscator_stream)

    @staticmethod
    def register_config_value_decryptor(config_value_decryptor):
        """Add a ConfigValueDecryptor that can be used to decrypt ConfigValues.

        Parameters
        ----------
        config_value_decryptor : ConfigValueDecryptor
            ConfigValueDecryptor to add to the list of decryptors that can be
            used to decrypt ConfigValues.

        Raises
        ------
        AssertionError
            If given config_value_decryptor is not a type of ConfigValueDecryptor
        """
        assert isinstance(config_value_decryptor, ConfigValueDecryptor)
        DecryptionUtils.__config_value_decryptors.append(config_value_decryptor)

    @staticmethod
    def create_and_register_config_value_decryptor(
        config_value_decryptor_implementer_name,
        config_value_decryptor_constructor_params=None
    ):
        """Attempts to dynamically create a ConfigValueDecryptor and register it.

        Parameters
        ----------
        config_value_decryptor_implementer_name : str
            Name of the ConfigValueDecryptor to dynamically load and create.
        config_value_decryptor_constructor_params : dict, optional
            Dictionary of constructor parameters to expand and pass to the ConfigValueDecryptor
            when it is created.

        Raises
        ------
        StepRunnerException
            If could not find class to load
            If loaded class is not a subclass of ConfigValueDecryptor

        See Also
        --------
        register_config_value_decryptor
        """
        if config_value_decryptor_constructor_params is None:
            config_value_decryptor_constructor_params = {}

        decryptor_class = DecryptionUtils.__get_decryption_class(
            config_value_decryptor_implementer_name
        )
        try:
            config_value_decryptor = decryptor_class(**config_value_decryptor_constructor_params)
            DecryptionUtils.register_config_value_decryptor(config_value_decryptor)
        except TypeError as error:
            raise ValueError(
                f"Loaded decryptor class ({decryptor_class}) failed to construct with given" +
                f" constructor arguments ({config_value_decryptor_constructor_params}): {error}"
            ) from error

    @staticmethod
    def decrypt(config_value):
        """If possible decrypt the given ConfigValue using one of the
        registered ConfigValueDecryptors.

        Parameters
        ----------
        config_value : ConfigValue
            ConfigValue to decrypt the value of with one of the
            registered ConfigValueDecryptors if possible.

        Returns
        -------
        obj or None
            Decrypted value of the given ConfigValue or
            None if none of the registered ConfigValueDecryptors can decrypt the given
            ConfigValue.
        """

        decrypted_value = None
        for config_value_decryptor in DecryptionUtils.__config_value_decryptors:
            if config_value_decryptor.can_decrypt(config_value):
                decrypted_value = config_value_decryptor.decrypt(config_value)
                break

        DecryptionUtils.__add_obfuscation_targets(decrypted_value)

        return decrypted_value

    @staticmethod
    def __add_obfuscation_targets(targets):
        if targets is not None:
            for obfuscator_stream in DecryptionUtils.__obfuscation_streams:
                obfuscator_stream.add_obfuscation_targets(targets)

    @staticmethod
    def __get_decryption_class(decryptor_implementer_name):
        """Given a decryptor implementer name dynamically loads the associated Class.

        Parameters
        ----------
        decryptor_implementer_name : str
            Either the short name of a ConfigValueDecryptor class which will be dynamically
            loaded from the 'ploigos_step_runner.config.decryptors' module or
            A class name that includes a dot seperated module name to load the Class from.

        Returns
        -------
        ConfigValueDecryptor
            Dynamically loaded subclass of ConfigValueDecryptor for given decryptor name.

        Raises
        ------
        StepRunnerException
            If could not find class to load
            If loaded class is not a subclass of ConfigValueDecryptor
        """
        parts = decryptor_implementer_name.split('.')
        class_name = parts.pop()
        module_name = '.'.join(parts)

        if not module_name:
            module_name = DecryptionUtils.__DEFAULT_DECRYPTORS_MODULE

        clazz = import_and_get_class(module_name, class_name)
        if not clazz:
            raise StepRunnerException(
                "Could not dynamically load decryptor implementer" +
                f" ({decryptor_implementer_name}) from module ({module_name})" +
                f" with class name ({class_name})"
            )
        if not issubclass(clazz, ConfigValueDecryptor):
            raise StepRunnerException(
                f"For decryptor implementer ({decryptor_implementer_name})" +
                f" dynamically loaded class ({clazz}) which is not sub class of" +
                f" ({ConfigValueDecryptor}) and should be."
            )

        return clazz



"""StepImplementer parent classes that are shared accross multiple steps.
"""

from ploigos_step_runner.step_implementers.shared.maven_generic import \
    MavenGeneric
from ploigos_step_runner.step_implementers.shared.maven_test_reporting_mixin import \
    MavenTestReportingMixin
from ploigos_step_runner.step_implementers.shared.npm_generic import NpmGeneric
from ploigos_step_runner.step_implementers.shared.openscap_generic import \
    OpenSCAPGeneric
from ploigos_step_runner.step_implementers.shared.rekor_sign_generic import \
    RekorSignGeneric



"""Shared utils for dealing with PGP keys.

Note that these functions use the "gpg" program to interact with PGP keys, but as PGP is the
standard and GPG is an implementation of that standard, we universally refer to anything having
to do with PGP keys as "PGP" rather then "GPG".
"""

import re
import sys
from io import StringIO
import sh

from ploigos_step_runner.utils.io import create_sh_redirect_to_multiple_streams_fn_callback

def detach_sign_with_pgp_key(file_to_sign_path, pgp_private_key_fingerprint, output_signature_path):
    """Does a detached sign of a given file using a given users private key.

    Parameters
    ----------
    file_to_sign_path : str
        Path to file to create a detached PGP signature for.

    pgp_private_key_fingerprint : str
        PGP private key fingerprint to sign the given file with.

    output_signature_path : str
        File path to put the detached signature in.

    Raises
    ------
    RuntimeError
        If error signing given file with given key.
    """
    try:
        sh.gpg( # pylint: disable=no-member
            '--armor',
            '--local-user',
            pgp_private_key_fingerprint,
            '--output',
            output_signature_path,
            '--detach-sign',
            file_to_sign_path
        )
    except sh.ErrorReturnCode as error:
        raise RuntimeError(
            f"Error performing detached signature of file ({file_to_sign_path})" \
            f" with PGP key ({pgp_private_key_fingerprint}): {error}"
        ) from error

def import_pgp_key(pgp_private_key):
    """Imports a PGP key.

    Parameters
    ----------
    pgp_private_key : str
        PGP key to import.

    Returns
    -------
    str
        Fingerprint of the imported PGP key.

    Raises
    ------
    RuntimeError
        If error getting PGP fingerprint for imported PGP key
        If error importing PGP key.
    """
    # Example input to match on:
    #   sec:-:3072:1:CF4AC14A3D109637:1601483310:1664555310::-:::scESC::::::23::0:
    #   fpr:::::::::DD7208BA0A6359F65B906B29CF4AC14A3D109637:
    #   grp:::::::::A483EE079EC1D58A954E3AAF3BCC61EDD7596BF0:
    gpg_regex = re.compile(r"^fpr:+([^:]+):$", re.MULTILINE)

    print("Import PGP private key to sign artifacts with")
    try:
        # import the key

        # NOTE: GPG is weird in that it sends "none error" output to stderr even on success...
        #       so merge the stderr into stdout
        gpg_import_stdout_result = StringIO()
        gpg_import_stdout_callback = create_sh_redirect_to_multiple_streams_fn_callback([
            sys.stdout,
            gpg_import_stdout_result
        ])
        sh.gpg( # pylint: disable=no-member
            '--import',
            '--fingerprint',
            '--with-colons',
            '--import-options=import-show',
            _in=pgp_private_key,
            _out=gpg_import_stdout_callback,
            _err_to_out=True,
            _tee='out'
        )

        # get the fingerprint of the imported key
        #
        # NOTE: if more then one match just using first one...
        gpg_imported_pgp_private_key_fingerprints = re.findall(
            gpg_regex,
            gpg_import_stdout_result.getvalue()
        )
        if len(gpg_imported_pgp_private_key_fingerprints) < 1:
            raise RuntimeError(
                "Error getting PGP fingerprint for PGP key"
                " to sign container image(s) with. See stdout and stderr for more info."
            )
        pgp_private_key_fingerprint = gpg_imported_pgp_private_key_fingerprints[0]

        print(
            "Imported PGP private key to sign artifacts with: "
            f"fingerprint='{pgp_private_key_fingerprint}'"
        )
    except sh.ErrorReturnCode as error:
        raise RuntimeError(
            f"Error importing pgp private key: {error}"
        ) from error

    return pgp_private_key_fingerprint

def export_pgp_public_key(pgp_private_key_fingerprint):
    """Exports a PGP public key given a private key fingerprint.

    Parameters
    ----------
    pgp_private_key_fingerprint : str
        PGP fingerprint.

    Returns
    -------
    str
        Public key from the private key fingerprint.

    Raises
    ------
    RuntimeError
        If error getting exported PGP public key
    """
    try:
        gpg_export_stdout_result = StringIO()

        sh.gpg( # pylint: disable=no-member
            '--armor',
            '--export',
            pgp_private_key_fingerprint,
            _out=gpg_export_stdout_result,
            _err_to_out=False,
            _tee='out'
        )

        gpg_public_key = gpg_export_stdout_result.getvalue()
    except sh.ErrorReturnCode as error:
        raise RuntimeError(
            f"Error exporting pgp public key: {error}"
        ) from error

    return gpg_public_key



"""Implementation of RekorGeneric in the Report step.

artifact_to_sign_uri_config_key is set to `results-archive-uri` as
this is the key used to obtain the artifact that needs to be
signed by rekor.

"""

from ploigos_step_runner.step_implementers.shared.rekor_sign_generic import RekorSignGeneric

class RekorSignReport(RekorSignGeneric):
    """ Rekor implementation specific to report
    """
    def __init__(  # pylint: disable=too-many-arguments
        self,
        workflow_result,
        parent_work_dir_path,
        config,
        environment=None
    ):
        super().__init__(
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path,
            config=config,
            environment=environment,
            artifact_to_sign_uri_config_key='results-archive-uri'
        )



"""Shared utils for dealing with dictionaries.
"""

def deep_merge(dest, source, overwrite_duplicate_keys=False, _path=None):
    """"deep merges source dictionary into destination dictionary.

    Parameters
    ----------
    dest : dict
        Destination dictionary to deep merge source into.
    source : dict
        Source dictionary to deep merge into dest.
    overwrite_duplicate_keys : bool
        True to overwite duplicate leaf keys in destination with source dictionary values.
        False to raise ValueError if any duplicate leaf values.

    Examples
    --------
    No Duplicate Keys:
    dict1 = {
        'A': {'value': '1'},
        'C': {'value': '1'}
    }
    dict2 = {
        'B': {'value': '2'},
    }
    expected_answer = {
        'A': {'value': '1'},
        'C': {'value': '1'},
        'B': {'value': '2'}
    }

    Duplicate keys (overwrite_duplicate_keys=True):

    dict1 = {
        'A': {'value': '1'},
        'C': {'value': 'overwriteme'}
    }
    dict2 = {
        'B': {'value': '2'},
        'C': {'value': '2'}
    }
    expected_answer = {
        'A': {'value': '1'},
        'C': {'value': '2'},
        'B': {'value': '2'}
    }

    Raises
    ------
    ValueError
        If source and destination contain a duplicate leaf key and overwrite_duplicate_keys is True

    Notes
    ------
    Modifies destination.

    Source is https://stackoverflow.com/questions/7204805/how-to-merge-dictionaries-of-dictionaries/7205107#7205107 # pylint: disable=line-too-long
    """
    if _path is None:
        _path = []

    for key in source:
        if key in dest:
            if isinstance(dest[key], dict) and isinstance(source[key], dict):
                deep_merge(
                    dest=dest[key],
                    source=source[key],
                    overwrite_duplicate_keys=overwrite_duplicate_keys,
                    _path=_path + [str(key)]
                )
            elif dest[key] == source[key]:
                pass # same leaf value
            else:
                if overwrite_duplicate_keys:
                    dest[key] = source[key]
                else:
                    raise ValueError('Conflict at %s' % '.'.join(_path + [str(key)]))
        else:
            dest[key] = source[key]
    return dest



"""`StepImplementer` for the `generate-metadata` to generate a semantic version from given input.

Supports the following semantic versions (https://semver.org/) :

  - major.minor.patch+build
  - major.minor.patch-pre_rleease+build

Notes
-----
When tagging container images we will regex + to - due
to https://github.com/docker/distribution/issues/1201.

Source for version sections:

  - major.minor.patch
    * will come from previous sub step of generate_metadata step,
      with step results including 'app_version'
    * known implementers:
      -  maven
  - pre-release
    * will come from previous sub step of generate_metadata step,
      with step results including 'pre_release'
    * known implementers:
      - git
  - build
    * will come from previous sub step of generate_metadata step,
      with step results including 'build'
    * known implementers:
      - git

Step Configuration
------------------
Step configuration expected as input to this step.
Could come from:

  * static configuration
  * runtime configuration
  * previous step results

Configuration Key | Required? | Default | Description
------------------|-----------|---------|-----------
`app-version`     | Yes       |         | Value to use for `version` portion of \
                                          semantic version (https://semver.org/).
`pre-release`     | No        |         | Value to use for `pre-release` portion \
                                          of semantic version (https://semver.org/). \
`build`           | Yes       |         | Value to use for `build` portion \
                                          of semantic version (https://semver.org/).

Result Artifacts
----------------
Results artifacts output by this step.

Result Artifact Key       | Description
--------------------------|------------
`version`                 | Constructed semantic version (https://semver.org/).
`container-image-version` | Constructed semantic version (https://semver.org/) without build number
"""

from ploigos_step_runner import StepImplementer
from ploigos_step_runner import StepResult

DEFAULT_CONFIG = {
    'release-branch': 'master'
}

REQUIRED_CONFIG_OR_PREVIOUS_STEP_RESULT_ARTIFACT_KEYS = [
    'app-version',
    'pre-release',
    'release-branch',
    'build'
]

class SemanticVersion(StepImplementer):  # pylint: disable=too-few-public-methods
    """`StepImplementer` for the `generate-metadata` to generate a
    semantic version from given input.
    """

    @staticmethod
    def step_implementer_config_defaults():
        """Getter for the StepImplementer's configuration defaults.

        Returns
        -------
        dict
            Default values to use for step configuration values.

        Notes
        -----
        These are the lowest precedence configuration values.

        """
        return DEFAULT_CONFIG

    @staticmethod
    def _required_config_or_result_keys():
        """Getter for step configuration or previous step result artifacts that are required before
        running this step.

        See Also
        --------
        _validate_required_config_or_previous_step_result_artifact_keys

        Returns
        -------
        array_list
            Array of configuration keys or previous step result artifacts
            that are required before running the step.
        """
        return REQUIRED_CONFIG_OR_PREVIOUS_STEP_RESULT_ARTIFACT_KEYS

    def _run_step(self):
        """Runs the step implemented by this StepImplementer.

        Returns
        -------
        StepResult
            Object containing the dictionary results of this step.
        """
        step_result = StepResult.from_step_implementer(self)

        app_version = None
        pre_release = None
        build = None
        release_branch = self.get_value('release-branch')
        app_version = self.get_value('app-version')
        pre_release = self.get_value('pre-release')
        build = self.get_value('build')

        if pre_release == release_branch:
            version = f'{app_version}+{build}'
            image_tag = f'{app_version}'
        else:
            version = f'{app_version}-{pre_release}+{build}'
            image_tag = f'{app_version}-{pre_release}'

        step_result.add_artifact(
            name='version',
            value=version
        )

        step_result.add_artifact(
            name='container-image-version',
            value=image_tag
        )

        step_result.add_evidence(
            name='version',
            value=version
        )
        step_result.add_evidence(
            name='container-image-version',
            value=image_tag
        )

        return step_result



import os
from io import IOBase
from unittest.mock import patch

import sh
from ploigos_step_runner import StepResult
from ploigos_step_runner.step_implementers.push_container_image import Skopeo
from testfixtures import TempDirectory
from tests.helpers.base_step_implementer_test_case import \
    BaseStepImplementerTestCase
from tests.helpers.test_utils import Any


class BaseTestStepImplementerSkopeoSourceBase(BaseStepImplementerTestCase):
    def create_step_implementer(
            self,
            step_config={},
            step_name='',
            implementer='',
            workflow_result=None,
            parent_work_dir_path=''
    ):
        return self.create_given_step_implementer(
            step_implementer=Skopeo,
            step_config=step_config,
            step_name=step_name,
            implementer=implementer,
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path
        )

class TestStepImplementerSkopeoSourceBase_step_implementer_config_defaults(
    BaseTestStepImplementerSkopeoSourceBase
):
    def test_result(self):
        defaults = Skopeo.step_implementer_config_defaults()
        expected_defaults = {
            'src-tls-verify': True,
            'dest-tls-verify': True,
            'container-image-pull-repository-type': 'containers-storage:',
            'container-image-push-repository-type': 'docker://'
        }
        self.assertEqual(defaults, expected_defaults)

class TestStepImplementerSkopeoSourceBase___required_config_or_result_keys(
    BaseTestStepImplementerSkopeoSourceBase
):
    def test_result(self):
        required_keys = Skopeo._required_config_or_result_keys()
        expected_required_keys = [
            'destination-url',
            ['source-tls,verify', 'src-tls-verify'],
            'dest-tls-verify',
            'service-name',
            'application-name',
            'organization',
            ['container-image-pull-tag', 'container-image-tag'],
            ['container-image-pull-repository-type', 'container-image-repository-type'],
            ['container-image-push-repository-type', 'container-image-repository-type']
        ]
        self.assertEqual(required_keys, expected_required_keys)

class TestStepImplementerSkopeoSourceBase__run_step(
    BaseTestStepImplementerSkopeoSourceBase
):
    def __run_pass_test(
        self,
        step_config,
        image_version,
        image_pull_tag,
        image_push_tag,
        temp_dir,
        skopeo_mock,
        skopeo_mock_call_dest_tls_value='true',
        skopeo_mock_call_src_tls_value='true',
        expected_step_result_message=None,
        containers_config_auth_file=None
    ):
        parent_work_dir_path = os.path.join(temp_dir.path, 'working')

        # setup step
        step_implementer = self.create_step_implementer(
            step_config=step_config,
            step_name='push-container-image',
            implementer='Skopeo',
            parent_work_dir_path=parent_work_dir_path,
        )

        # run step
        result = step_implementer._run_step()

        # verify
        expected_step_result = StepResult(
            step_name='push-container-image',
            sub_step_name='Skopeo',
            sub_step_implementer_name='Skopeo'
        )
        if expected_step_result_message:
            expected_step_result.message = expected_step_result_message
        expected_step_result.add_artifact(
            name='container-image-registry-uri',
            value='fake-registry.xyz',
            description='Registry URI poriton of the container image tag' \
                ' of the pushed container image.'
        )
        expected_step_result.add_artifact(
            name='container-image-registry-organization',
            value='fake-org',
            description='Organization portion of the container image tag' \
                ' of the pushed container image.'
        )
        expected_step_result.add_artifact(
            name='container-image-repository',
            value='fake-app-fake-service',
            description='Repository portion of the container image tag' \
                ' of the pushed container image.'
        )
        expected_step_result.add_artifact(
            name='container-image-name',
            value='fake-app-fake-service',
            description='Another way to reference the' \
                ' repository portion of the container image tag of the pushed container image.'
        )
        expected_step_result.add_artifact(
            name='container-image-version',
            value=image_version,
            description='Version portion of the container image tag of the pushed container image.'
        )
        expected_step_result.add_artifact(
            name='container-image-tag',
            value=f'fake-registry.xyz/fake-org/fake-app-fake-service:{image_version}',
            description='Full container image tag of the pushed container,' \
                ' including the registry URI.'
        )
        expected_step_result.add_artifact(
            name='container-image-short-tag',
            value=f'fake-org/fake-app-fake-service:{image_version}',
            description='Short container image tag of the pushed container image,' \
                ' excluding the registry URI.'
        )
        self.assertEqual(result, expected_step_result)

        if not containers_config_auth_file:
            containers_config_auth_file = os.path.join(
                step_implementer.work_dir_path,
                'container-auth.json'
            )

        skopeo_mock.copy.assert_called_once_with(
            f"--src-tls-verify={skopeo_mock_call_src_tls_value}",
            f"--dest-tls-verify={skopeo_mock_call_dest_tls_value}",
            f"--authfile={containers_config_auth_file}",
            f'containers-storage:{image_pull_tag}',
            f'docker://{image_push_tag}',
            _out=Any(IOBase),
            _err=Any(IOBase),
            _tee='err'
        )

    @patch.object(sh, 'skopeo', create=True)
    def test_pass(self, skopeo_mock):
        with TempDirectory() as temp_dir:
            image_version = '1.0-69442c8'
            image_pull_tag = f'localhost/fake-org/fake-app-fake-service:{image_version}'
            image_push_tag = f'fake-registry.xyz/fake-org/fake-app-fake-service:{image_version}'
            step_config = {
                'destination-url': 'fake-registry.xyz',
                'service-name': 'fake-service',
                'application-name': 'fake-app',
                'organization': 'fake-org',
                'container-image-version': image_version,
                'container-image-pull-tag': image_pull_tag
            }
            self.__run_pass_test(
                temp_dir=temp_dir,
                image_version=image_version,
                image_pull_tag=image_pull_tag,
                image_push_tag=image_push_tag,
                step_config=step_config,
                skopeo_mock=skopeo_mock
            )

    @patch.object(sh, 'skopeo', create=True)
    def test_pass_default_version(self, skopeo_mock):
        with TempDirectory() as temp_dir:
            image_version = 'latest'
            image_pull_tag = f'localhost/fake-org/fake-app-fake-service:{image_version}'
            image_push_tag = f'fake-registry.xyz/fake-org/fake-app-fake-service:{image_version}'
            step_config = {
                'destination-url': 'fake-registry.xyz',
                'service-name': 'fake-service',
                'application-name': 'fake-app',
                'organization': 'fake-org',
                'container-image-pull-tag': image_pull_tag
            }
            self.__run_pass_test(
                temp_dir=temp_dir,
                image_version=image_version,
                image_pull_tag=image_pull_tag,
                image_push_tag=image_push_tag,
                step_config=step_config,
                skopeo_mock=skopeo_mock
            )


    @patch.object(sh, 'skopeo', create=True)
    def test_pass_string_destination_tls_truethy(self, skopeo_mock):
        with TempDirectory() as temp_dir:
            image_version = '1.0-69442c8'
            image_pull_tag = f'localhost/fake-org/fake-app-fake-service:{image_version}'
            image_push_tag = f'fake-registry.xyz/fake-org/fake-app-fake-service:{image_version}'
            step_config = {
                'destination-url': 'fake-registry.xyz',
                'service-name': 'fake-service',
                'application-name': 'fake-app',
                'organization': 'fake-org',
                'container-image-version': image_version,
                'container-image-pull-tag': image_pull_tag,
                'dest-tls-verify': 'true'
            }
            self.__run_pass_test(
                temp_dir=temp_dir,
                image_version=image_version,
                image_pull_tag=image_pull_tag,
                image_push_tag=image_push_tag,
                step_config=step_config,
                skopeo_mock=skopeo_mock
            )

    @patch.object(sh, 'skopeo', create=True)
    def test_pass_string_destination_tls_falsy(self, skopeo_mock):
        with TempDirectory() as temp_dir:
            image_version = '1.0-69442c8'
            image_pull_tag = f'localhost/fake-org/fake-app-fake-service:{image_version}'
            image_push_tag = f'fake-registry.xyz/fake-org/fake-app-fake-service:{image_version}'
            step_config = {
                'destination-url': 'fake-registry.xyz',
                'service-name': 'fake-service',
                'application-name': 'fake-app',
                'organization': 'fake-org',
                'container-image-version': image_version,
                'container-image-pull-tag': image_pull_tag,
                'dest-tls-verify': 'false'
            }
            self.__run_pass_test(
                temp_dir=temp_dir,
                image_version=image_version,
                image_pull_tag=image_pull_tag,
                image_push_tag=image_push_tag,
                step_config=step_config,
                skopeo_mock=skopeo_mock,
                skopeo_mock_call_dest_tls_value='false'
            )

    @patch.object(sh, 'skopeo', create=True)
    def test_pass_string_source_tls_truthy(self, skopeo_mock):
        with TempDirectory() as temp_dir:
            image_version = '1.0-69442c8'
            image_pull_tag = f'localhost/fake-org/fake-app-fake-service:{image_version}'
            image_push_tag = f'fake-registry.xyz/fake-org/fake-app-fake-service:{image_version}'
            step_config = {
                'destination-url': 'fake-registry.xyz',
                'service-name': 'fake-service',
                'application-name': 'fake-app',
                'organization': 'fake-org',
                'container-image-version': image_version,
                'container-image-pull-tag': image_pull_tag,
                'src-tls-verify': 'true'
            }
            self.__run_pass_test(
                temp_dir=temp_dir,
                image_version=image_version,
                image_pull_tag=image_pull_tag,
                image_push_tag=image_push_tag,
                step_config=step_config,
                skopeo_mock=skopeo_mock
            )

    @patch.object(sh, 'skopeo', create=True)
    def test_pass_string_source_tls_falsy(self, skopeo_mock):
        with TempDirectory() as temp_dir:
            image_version = '1.0-69442c8'
            image_pull_tag = f'localhost/fake-org/fake-app-fake-service:{image_version}'
            image_push_tag = f'fake-registry.xyz/fake-org/fake-app-fake-service:{image_version}'
            step_config = {
                'destination-url': 'fake-registry.xyz',
                'service-name': 'fake-service',
                'application-name': 'fake-app',
                'organization': 'fake-org',
                'container-image-version': image_version,
                'container-image-pull-tag': image_pull_tag,
                'src-tls-verify': 'false'
            }
            self.__run_pass_test(
                temp_dir=temp_dir,
                image_version=image_version,
                image_pull_tag=image_pull_tag,
                image_push_tag=image_push_tag,
                step_config=step_config,
                skopeo_mock=skopeo_mock,
                skopeo_mock_call_src_tls_value='false'
            )

    @patch.object(sh, 'skopeo', create=True)
    def test_pass_custom_auth_file(self, skopeo_mock):
        with TempDirectory() as temp_dir:
            image_version = '1.0-69442c8'
            image_pull_tag = f'localhost/fake-org/fake-app-fake-service:{image_version}'
            image_push_tag = f'fake-registry.xyz/fake-org/fake-app-fake-service:{image_version}'
            step_config = {
                'destination-url': 'fake-registry.xyz',
                'service-name': 'fake-service',
                'application-name': 'fake-app',
                'organization': 'fake-org',
                'container-image-version': image_version,
                'container-image-pull-tag': image_pull_tag,
                'containers-config-auth-file': 'mock-auth.json'
            }
            self.__run_pass_test(
                temp_dir=temp_dir,
                image_version=image_version,
                image_pull_tag=image_pull_tag,
                image_push_tag=image_push_tag,
                step_config=step_config,
                skopeo_mock=skopeo_mock,
                containers_config_auth_file='mock-auth.json'
            )

    @patch.object(sh, 'skopeo', create=True)
    def test_fail_run_skopeo(self, skopeo_mock):
        with TempDirectory() as temp_dir:
            parent_work_dir_path = os.path.join(temp_dir.path, 'working')

            # setup step
            image_version = '1.0-69442c8'
            image_pull_tag = f'localhost/fake-org/fake-app-fake-service:{image_version}'
            image_push_tag = f'fake-registry.xyz/fake-org/fake-app-fake-service:{image_version}'
            step_config = {
                'destination-url': 'fake-registry.xyz',
                'service-name': 'fake-service',
                'application-name': 'fake-app',
                'organization': 'fake-org',
                'container-image-version': image_version,
                'container-image-version': image_version,
                'container-image-pull-tag': image_pull_tag
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                step_name='push-container-image',
                implementer='Skopeo',
                parent_work_dir_path=parent_work_dir_path,
            )

            # run step (mock fail)
            skopeo_mock.copy.side_effect = sh.ErrorReturnCode('skopeo', b'mock stdout', b'mock error')
            result = step_implementer._run_step()

            # verify
            expected_step_result = StepResult(
                step_name='push-container-image',
                sub_step_name='Skopeo',
                sub_step_implementer_name='Skopeo'
            )
            expected_step_result.add_artifact(
                name='container-image-registry-uri',
                value='fake-registry.xyz',
                description='Registry URI poriton of the container image tag' \
                    ' of the pushed container image.'
            )
            expected_step_result.add_artifact(
                name='container-image-registry-organization',
                value='fake-org',
                description='Organization portion of the container image tag' \
                    ' of the pushed container image.'
            )
            expected_step_result.add_artifact(
                name='container-image-repository',
                value='fake-app-fake-service',
                description='Repository portion of the container image tag' \
                    ' of the pushed container image.'
            )
            expected_step_result.add_artifact(
                name='container-image-name',
                value='fake-app-fake-service',
                description='Another way to reference the' \
                    ' repository portion of the container image tag of the pushed container image.'
            )
            expected_step_result.add_artifact(
                name='container-image-version',
                value=image_version,
                description='Version portion of the container image tag of the pushed container image.'
            )
            expected_step_result.add_artifact(
                name='container-image-tag',
                value='fake-registry.xyz/fake-org/fake-app-fake-service:1.0-69442c8',
                description='Full container image tag of the pushed container,' \
                    ' including the registry URI.'
            )
            expected_step_result.add_artifact(
                name='container-image-short-tag',
                value='fake-org/fake-app-fake-service:1.0-69442c8',
                description='Short container image tag of the pushed container image,' \
                    ' excluding the registry URI.'
            )
            expected_step_result.success = False
            expected_step_result.message = f"Error pushing container image ({image_pull_tag}) " +\
                f" to tag ({image_push_tag}) using skopeo: \n" +\
                f"\n" +\
                f"  RAN: skopeo\n" +\
                f"\n" +\
                f"  STDOUT:\n" +\
                f"mock stdout\n" +\
                f"\n" +\
                f"  STDERR:\n" +\
                f"mock error"
            self.assertEqual(result, expected_step_result)

            containers_config_auth_file = os.path.join(
                step_implementer.work_dir_path,
                'container-auth.json'
            )
            skopeo_mock.copy.assert_called_once_with(
                "--src-tls-verify=true",
                "--dest-tls-verify=true",
                f"--authfile={containers_config_auth_file}",
                f'containers-storage:{image_pull_tag}',
                f'docker://{image_push_tag}',
                _out=Any(IOBase),
                _err=Any(IOBase),
                _tee='err'
            )



import os

import unittest
from testfixtures import TempDirectory

from tests.helpers.base_test_case import BaseTestCase

from ploigos_step_runner.utils.reflection import import_and_get_class

class TestReflectionUtils(BaseTestCase):
    def test_import_and_get_class_module_does_not_exist(self):
        self.assertIsNone(
            import_and_get_class('does.not.exist', 'HelloWorld')
        )

    def test_import_and_get_class_class_does_not_exist(self):
        self.assertIsNone(
            import_and_get_class('ploigos_step_runner.step_implementers.package', 'HelloWorld')
        )

    def test_import_and_get_class_class_exists_in_module(self):
        self.assertIsNotNone(
            import_and_get_class('ploigos_step_runner.step_implementers.container_image_static_compliance_scan', 'OpenSCAP')
        )



"""Representation of an individual step's step configuration.
"""

import copy

from ploigos_step_runner.config.sub_step_config import SubStepConfig


class StepConfig:
    """Representation of an individual step's step configuration.

    Parameters
    ----------
    parent_config : Config
        Parent configuration containing this step configuration.
    step_name : str
        Name of the step.

    Attributes
    ----------
    __parent_config : Config
    __step_name : str
    __sub_steps : list of SubStepConfig
    __sub_step_config_overrides : dict
    """

    def __init__(self, parent_config, step_name):
        self.__parent_config = parent_config
        self.__step_name = step_name
        self.__sub_steps = []
        self.__step_config_overrides = {}

    @property
    def parent_config(self):
        """
        Returns
        -------
        Config
        """
        return self.__parent_config

    @property
    def step_name(self):
        """
        Returns
        -------
        str
            Name of this step.
        """
        return self.__step_name

    @property
    def sub_steps(self):
        """
        Returns
        -------
        list of SubStepConfig
            Sub steps of this step.
        """
        return self.__sub_steps

    def get_sub_step(self, sub_step_name):
        """Get sub step of this step with a given name if one exists.

        Parameters
        ----------
        sub_step_name : str
            Name of the sub step of this step to get.

        Returns
        -------
        SubStepConfig
            Sub step of this step with the given name or
            None if no sub step with given name exists as part of this step.
        """
        for sub_step in self.sub_steps:
            if sub_step_name == sub_step.sub_step_name:
                return sub_step

        return None

    @property
    def step_config_overrides(self):
        """Gets a deep copy of the step configuration overrides.

        Returns
        -------
        dict
            Deep copy of the step configuration overrides.
        """
        return copy.deepcopy(self.__step_config_overrides)

    @step_config_overrides.setter
    def step_config_overrides(self, step_config_overrides):
        """Sets the step configuration overrides.

        Parameters
        ----------
        step_config_overrides : dict
            New step configuration overrides.
        """
        self.__step_config_overrides = step_config_overrides if step_config_overrides else {}

    def add_or_update_sub_step_config(
        self,
        sub_step_name,
        sub_step_implementer_name,
        sub_step_config_dict=None,
        sub_step_env_config=None,
        sub_step_contine_sub_steps_on_failure=False
    ): # pylint: disable=too-many-arguments
        """Add a new or update an existing sub step configuration for this step.

        Parameters
        ----------
        sub_step_name : str
            Name of the sub step to add or update.
        sub_step_implementer_name : str
            Name of the sub step implementer for the sub step being added or updated.
            If updating this can not be different then existing sub step with the same name.
        sub_step_config_dict : dict, optional
            Sub step configuration to add or update for named sub step.
            If updating this can not have any duplicative leaf keys to the existing
                sub step configuration.
        sub_step_env_config : dict, optional
            Sub step environment configuration to add or update for named sub step.
            If updating this can not have any duplicative leaf keys to the existing
                sub step environment configuration.
        sub_step_contine_sub_steps_on_failure : bool
            True to continue executing other sub steps in current step if this sub step fails.
            False to fail all step execution if this sub step fails.

        Raises
        ------
        AssertionError
            If attempt to update an existing sub step and new and existing sub step implementers
                do not match.
        ValueError
            If new sub step configuration has duplicative leaf keys to
                existing sub step configuration.
            If new sub step environment configuration has duplicative leaf keys to
                existing sub step environment configuration.
        """

        existing_sub_step_config = None
        for sub_step_config in self.sub_steps:
            if sub_step_config.sub_step_name == sub_step_name:
                existing_sub_step_config = sub_step_config
                break

        if existing_sub_step_config is None:
            sub_step_config = SubStepConfig(
                parent_step_config=self,
                sub_step_name=sub_step_name,
                sub_step_implementer_name=sub_step_implementer_name,
                sub_step_config_dict=sub_step_config_dict,
                sub_step_env_config=sub_step_env_config,
                sub_step_contine_sub_steps_on_failure=sub_step_contine_sub_steps_on_failure
            )
            self.sub_steps.append(sub_step_config)
        else:
            assert sub_step_implementer_name == existing_sub_step_config.sub_step_implementer_name,\
                f"Step ({self.step_name}) failed to update sub step ({sub_step_name})" + \
                " with new config due to new sub step implementer" + \
                f" ({sub_step_implementer_name}) not matching existing sub step implementer" + \
                f" ({existing_sub_step_config.sub_step_implementer_name})."

            assert sub_step_contine_sub_steps_on_failure == \
                    existing_sub_step_config.sub_step_contine_sub_steps_on_failure, \
                f"Step ({self.step_name}) failed to update sub step ({sub_step_name})" + \
                " with new config due to new continue sub steps on failure" + \
                f" ({sub_step_contine_sub_steps_on_failure}) not matching existing" + \
                " continue sub steps on failure " + \
                f" ({existing_sub_step_config.sub_step_implementer_name})."

            existing_sub_step_config.merge_sub_step_config(sub_step_config_dict)
            existing_sub_step_config.merge_sub_step_env_config(sub_step_env_config)



"""`StepImplementer` for the `generate-metadata` step using Commitizen.

Step Configuration
------------------
Step configuration expected as input to this step.
Could come from:

  * static configuration
  * runtime configuration
  * previous step results

Configuration Key  | Required? | Default    | Description
-------------------|-----------|------------|-----------
`repo-root`        | Yes       | `./`       | Directory path to the Git repo to anaylze tags.
`cz-json`          | Yes       | `.cz.json` | json file with the commitizen configurataion.

Result Artifacts
----------------
Results artifacts output by this step.

Result Artifact Key | Description
--------------------|------------
`app-version`       | Value to use for `version` portion of semantic version \
                      (https://semver.org/). Uses the version from `cz bump` in dry-run mode.
"""

import json
import io
import os
import re
import sys

import sh

from git import Repo
from ploigos_step_runner import StepImplementer, StepResult

DEFAULT_CONFIG = {
    'cz-json': '.cz.json',
    'repo-root': './',
}

REQUIRED_CONFIG_OR_PREVIOUS_STEP_RESULT_ARTIFACT_KEYS = [
    'cz-json',
    'repo-root'
]

class Commitizen(StepImplementer):
    """
    StepImplementer for the generate-metadata step for Commitizen.
    """

    @staticmethod
    def step_implementer_config_defaults():
        """Getter for the StepImplementer's configuration defaults.

        Returns
        -------
        dict
            Default values to use for step configuration values.

        Notes
        -----
        These are the lowest precedence configuration values.

        """
        return DEFAULT_CONFIG

    @staticmethod
    def _required_config_or_result_keys():
        """Getter for step configuration or previous step result artifacts that are required before
        running this step.

        See Also
        --------
        _validate_required_config_or_previous_step_result_artifact_keys

        Returns
        -------
        array_list
            Array of configuration keys or previous step result artifacts
            that are required before running the step.
        """
        return REQUIRED_CONFIG_OR_PREVIOUS_STEP_RESULT_ARTIFACT_KEYS

    def _validate_required_config_or_previous_step_result_artifact_keys(self):
        """Validates that the required configuration keys or previous step result artifacts
        are set and have valid values.

        Validates that:
        * required configuration is given
        * given 'cz-json' exists

        Raises
        ------
        AssertionError
            If step configuration or previous step result artifacts have invalid required values
        """
        super()._validate_required_config_or_previous_step_result_artifact_keys()

        cz_json_path = self.get_value('cz-json')
        assert os.path.exists(cz_json_path), \
            f'cz-json does not exist: {cz_json_path}'

        with open(cz_json_path, 'rb') as cz_json:
            assert json.loads(cz_json.read()), "cz-json is not valid JSON"

    def _run_step(self):
        """Runs the step implemented by this StepImplementer.

        Returns
        -------
        StepResult
            Object containing the dictionary results of this step.
        """

        step_result = StepResult.from_step_implementer(self)
        cz_json_path = self.get_value('cz-json')

        repo_root = self.get_value('repo-root')
        repo = Repo(repo_root)
        os.chdir(repo_root)

        with open(cz_json_path, 'rb+') as cz_json:
            cz_json_contents = json.loads(cz_json.read())
            cz_json_contents['commitizen']['version'] = self._get_version_tag(repo.tags)
            cz_json.seek(0)
            cz_json.truncate(0)
            cz_json.write(json.dumps(cz_json_contents).encode())

        out = io.StringIO()
        sh.cz.bump( # pylint: disable=no-member
            '--dry-run',
            '--yes',
            _out=out,
            _err=sys.stderr,
            _tee='err'
        )
        bump_regex = r'tag to create: (\d+.\d+.\d+)'
        version = re.findall(bump_regex, out.getvalue(),)[0]
        step_result.add_artifact(name='app-version', value=version)

        return step_result

    @staticmethod
    def _get_version_tag(tags):
        """Gets the latest tag version and updates the json config.

        Returns
        -------
        tag_version
            String representing the latest repo tag version.
        """

        tag_version = '0.0.0'

        # check for existing repo tags. if no existing repo tag is found then default to version
        # 0.0.0. if existing repo tags are found then use the latest version tag.
        for tag in tags:
            semantic_version = re.match(r'.*(\d+).(\d+).(\d+).*', tag.name)
            for i, part in enumerate(semantic_version.groups()):
                tag_version_part = tag_version.split('.')[i]
                if part > tag_version_part:
                    tag_version = ".".join(semantic_version.groups())
                elif part < tag_version_part:
                    break

        return tag_version



"""Test StepResultEvidence
"""

from ploigos_step_runner.results import StepResultEvidence
from tests.helpers.base_test_case import BaseTestCase


class TestStepResultEvidenceTest(BaseTestCase):
    """Test StepResultEvidence
    """

    def test_name(self):
        evidence = StepResultEvidence(
            name='foo',
            value='hello'
        )

        self.assertEqual('foo', evidence.name)

    def test_value(self):
        evidence = StepResultEvidence(
            name='foo',
            value='hello'
        )

        self.assertEqual('hello', evidence.value)

    def test_description_empty(self):
        evidence = StepResultEvidence(
            name='foo',
            value='hello'
        )

        self.assertEqual("", evidence.description)

    def test_description(self):
        evidence = StepResultEvidence(
            name='foo',
            value='hello',
            description='test description'
        )

        self.assertEqual("test description", evidence.description)

    def test_as_dict_no_description(self):
        evidence = StepResultEvidence(
            name='foo',
            value='hello'
        )

        expected = {'description': '', 'name': 'foo', 'value': 'hello'}
        self.assertEqual(expected, evidence.as_dict())

    def test_as_dict_with_description(self):
        evidence = StepResultEvidence(
            name='foo',
            value='hello',
            description='test description'
        )

        expected = {'description': 'test description', 'name': 'foo', 'value': 'hello'}
        self.assertEqual(expected, evidence.as_dict())

    def test___str__(self):
        evidence = StepResultEvidence(
            name='foo',
            value='hello',
            description='test description'
        )

        expected = "{'name': 'foo', 'value': 'hello', 'description': 'test description'}"
        self.assertEqual(expected, str(evidence))

    def test___repr__(self):
        evidence = StepResultEvidence(
            name='foo',
            value='hello',
            description='test description'
        )

        expected = "StepResultEvidence(name=foo, value=hello, description=test description)"
        self.assertEqual(expected, repr(evidence))

    def test___eq__(self):
        evidence1 = StepResultEvidence(
            name='foo',
            value='hello',
            description='test description'
        )

        evidence2 = StepResultEvidence(
            name='foo',
            value='hello',
            description='test description'
        )

        self.assertEqual(evidence1, evidence2)

    def test___nq__name(self):
        evidence1 = StepResultEvidence(
            name='foo',
            value='hello',
            description='test description'
        )

        evidence2 = StepResultEvidence(
            name='bad',
            value='hello',
            description='test description'
        )

        self.assertNotEqual(evidence1, evidence2)

    def test___nq__value(self):
        evidence1 = StepResultEvidence(
            name='foo',
            value='hello',
            description='test description'
        )

        evidence2 = StepResultEvidence(
            name='foo',
            value='bad',
            description='test description'
        )

        self.assertNotEqual(evidence1, evidence2)

    def test___nq__description(self):
        evidence1 = StepResultEvidence(
            name='foo',
            value='hello',
            description='test description'
        )

        evidence2 = StepResultEvidence(
            name='foo',
            value='hello',
            description='bad'
        )

        self.assertNotEqual(evidence1, evidence2)



"""Abstract class and helper constants for WorkflowResult
"""
import json
import os
import pickle

import yaml
from ploigos_step_runner.exceptions import StepRunnerException
from ploigos_step_runner.results import StepResult
from ploigos_step_runner.utils.dict import deep_merge
from ploigos_step_runner.utils.file import create_parent_dir


class WorkflowResult:
    """
    Class to manage a list of StepResults.
    The WorkflowResult represents ALL previous results.
    """

    def __init__(self):
        self.__workflow_list = []

    @property
    def workflow_list(self):
        """Return workflow_list
        """
        return self.__workflow_list

    def get_artifact_value(
        self,
        artifact,
        step_name=None,
        sub_step_name=None,
        environment=None
    ): # pylint: disable=too-many-boolean-expressions
        """Search for an artifact.

        If step_name, sub_step_name, or environment are not provided ensure the artifact comes
        from the last step that returned that artifact.

        1.  if step_name is provided, look for the artifact in the step
        2.  elif step_name and sub_step_name is provided, look for the artifact in the step/sub_step
        3.  else, search ALL steps for the FIRST match of the artifact.

        Parameters
        ----------
        artifact: str
           The artifact name to search for
        step_name: str optional
           Optionally search only in one step
        sub_step_name: str optional
            Optionally search only in one step
        environment : str
            Optional. Environment to get the step result for.

        Returns
        -------
        Str
           'v1.0.2'
        """

        value = None
        for step_result in reversed(self.workflow_list):
            if ( \
                (not step_name or step_result.step_name == step_name) and \
                (not sub_step_name or step_result.sub_step_name == sub_step_name) and \
                (not environment or step_result.environment == environment)
            ):
                value = step_result.get_artifact_value(name=artifact)
                if value is not None:
                    break

        return value

    def get_evidence_value(
        self,
        evidence,
        step_name=None,
        sub_step_name=None,
        environment=None
    ): # pylint: disable=too-many-boolean-expressions
        """Search for an evidence.

        If step_name, sub_step_name, or environment are provided ensure the evidence comes
        from the first

        1.  if step_name is provided, look for the evidence in the step
        2.  elif step_name and sub_step_name is provided, look for the evidence in the step/sub_step
        3.  else, search ALL steps for the FIRST match of the evidence.

        Parameters
        ----------
        evidence: str
           The artifact name to search for
        step_name: str optional
           Optionally search only in one step
        sub_step_name: str optional
            Optionally search only in one step
        environment : str
            Optional. Environment to get the step result for.

        Returns
        -------
        Str
           'v1.0.2'
        """

        value = None
        for step_result in self.workflow_list:
            if ( \
                (not step_name or step_result.step_name == step_name) and \
                (not sub_step_name or step_result.sub_step_name == sub_step_name) and \
                (not environment or step_result.environment == environment)
            ):
                value = step_result.get_evidence_value(name=evidence)
                if value is not None:
                    break

        return value

    def add_step_result(self, step_result):
        """Add a single step_result to the workflow list.

        If the new result step is not already in the list
           - simply append, done
        Else
           - find the old step result
           - merge the old artifacts into the new artifacts
           - delete the old step result
           - append the new step result
           - note: the delete/append is needed because it is a list

        Parameters
        ----------
        step_result : StepResult
           An StepResult object to add to the list

        Raises
        ------
        Raises a StepRunnerException if an instance other than
        StepResult is passed as a parameter
        """

        if isinstance(step_result, StepResult):
            if self.__step_result_exists(step_result):
                raise StepRunnerException(
                    f'Can not add duplicate StepResult for step ({step_result.step_name}),'
                    f' sub step ({step_result.sub_step_name}),'
                    f' and environment ({step_result.environment}).'
                )

            self.workflow_list.append(step_result)

        else:
            raise StepRunnerException('expect StepResult instance type')

    # ARTIFACT helpers:
    def write_results_to_yml_file(self, yml_filename):
        """Write the workflow list in a yaml format to file

        Parameters
        ----------
        yml_filename : str
             Name of file to write (eg: step-runner-results/step-runner-results.yml)

        Raises
        ------
        Raises a RuntimeError if the file cannot be dumped
        """
        try:
            create_parent_dir(yml_filename)
            with open(yml_filename, 'w', encoding='utf-8') as file:
                results = self.__get_all_step_results_dict()
                yaml.dump(results, file, indent=4)
        except Exception as error:
            raise RuntimeError(f'error dumping {yml_filename}: {error}') from error

    def write_results_to_json_file(self, json_filename):
        """Write the workflow list in a json format to file.

        Parameters
        ----------
        json_filename : str
             Name of file to write (eg: step-runner-results.json)

        Raises
        ------
        Raises a RuntimeError if the file cannot be dumped
        """
        try:
            create_parent_dir(json_filename)
            with open(json_filename, 'w', encoding='utf-8') as file:
                results = self.__get_all_step_results_dict()
                json.dump(results, file, indent=4)
        except Exception as error:
            raise RuntimeError(f'error dumping {json_filename}: {error}') from error

    # File handlers

    @staticmethod
    def load_from_pickle_file(pickle_filename):
        """Return the contents of a pickled file.

        The file is expected to contain WorkflowResult instances

        Parameters
        ----------
        pickle_filename: str
           Name of the file to load

        Raises
        ------
        Raises a StepRunnerException if the file cannot be loaded
        Raises a StepRunnerException if the file contains non WorkflowResult instances
        """
        try:
            create_parent_dir(pickle_filename)

            # if the file does not exist return empty object
            if not os.path.isfile(pickle_filename):
                return WorkflowResult()

            # if the file is empty return empty object
            if os.path.getsize(pickle_filename) == 0:
                return WorkflowResult()

            # check that the file has Workflow object
            with open(pickle_filename, 'rb') as file:
                workflow_result = pickle.load(file)
                if not isinstance(workflow_result, WorkflowResult):
                    raise StepRunnerException(f'error {pickle_filename} has invalid data')
                return workflow_result

        except Exception as error:
            raise StepRunnerException(f'error loading {pickle_filename}: {error}') from error

    def merge_with_pickle_file(self, pickle_filename):
        """Merge our workflow list with that stored on disk.
        When we find overlaps, our in-memory values win.

        Note: any locking of the pickle file is the responsibility
        of the caller.

        Parameters
        ----------
        pickle_filename : the on-disk path to the pickle file to merge with.

        """
        on_disk_wfr = WorkflowResult.load_from_pickle_file(pickle_filename)
        on_disk_results = on_disk_wfr.workflow_list
        merged_workflow_list = []

        for in_mem_step_result in self.__workflow_list:
            on_disk_step_result = on_disk_wfr.get_step_result(
                step_name=in_mem_step_result.step_name,
                sub_step_name=in_mem_step_result.sub_step_name,
                environment=in_mem_step_result.environment
            )

            if on_disk_step_result:
                on_disk_results.remove(on_disk_step_result)

                # in-memory values win if the two results are unequal
                if on_disk_step_result != in_mem_step_result:
                    on_disk_step_result.merge(in_mem_step_result)

                merged_workflow_list.append(on_disk_step_result)

            else:
                merged_workflow_list.append(in_mem_step_result)

        # on_disk_results at this point will only have the step results
        # that were on disk, but not in memory.
        merged_workflow_list += on_disk_results

        self.__workflow_list = merged_workflow_list

    def write_to_pickle_file(self, pickle_filename):
        """Write the workflow list in a pickle format to file.

        Note: any locking of the pickle file is the responsibility
        of the caller.

        Parameters
        ----------
        pickle_filename : str
             Name of file to write (eg: step-runner-results.pkl)

        Raises
        ------
        Raises a RuntimeError if the file cannot be dumped
        """
        try:
            create_parent_dir(pickle_filename)
            with open(pickle_filename, 'wb') as file:
                pickle.dump(self, file)
        except Exception as error:
            raise RuntimeError(f'error dumping {pickle_filename}: {error}') from error

    def __step_result_exists(self, step_result):
        """Return True if the provided StepResult exists in our
        workflow list, False otherwise.

        Parameters
        ----------
        step_result - a StepResult object to search for

        Return value
        ------------
            True if StepResult exists, False otherwise
        """
        existing_step_result = self.get_step_result(
            step_name=step_result.step_name,
            sub_step_name=step_result.sub_step_name,
            environment=step_result.environment
        )
        return existing_step_result

    def __get_all_step_results_dict(self):
        """Get a dictionary of all of the recorded StepResults.

        Returns
        -------
        results: dict
            results of all steps from list
        """
        all_results = {}
        for step_result in self.workflow_list:
            all_results = deep_merge(
                dest=all_results,
                source=step_result.get_step_result_dict(),
                overwrite_duplicate_keys=True
            )
        step_runner_results = {
            'step-runner-results': all_results
        }
        return step_runner_results

    def get_step_result(
        self,
        step_name,
        sub_step_name=None,
        environment=None
    ): # pylint: disable=too-many-boolean-expressions
        """Helper method to return a step result.

        Parameters
        ----------
        step_name: str
            Name of step to search for
        sub_step_name: str
            Name of sub step to search for
        environment : str
            Optional. Environment to get the step result for.

        Returns
        -------
        StepResult
        """

        for step_result in self.workflow_list:
            if ( \
                (not step_name or step_result.step_name == step_name) and \
                (not sub_step_name or step_result.sub_step_name == sub_step_name) and \
                (not environment or step_result.environment == environment)
            ):
                return step_result

        return None



"""`StepImplementer` for the `container-image-static-vulnerability-scan` step using OpenSCAP.

Convenience wrapper for the `OpenSCAPGeneric` step implementer so that users
can specify `OpenSCAP` for the `container-image-static-compliance-scan`.

Step Configuration
------------------
Step configuration expected as input to this step.
Could come from:

  * static configuration
  * runtime configuration
  * previous step results

 Configuration Key             | Required? | Default | Description
-------------------------------|-----------|---------|-----------
`container-image-tag`          | Yes       |         | Container image tag to scan.
`oscap-input-definitions-uri`  | Yes       |         | URI to the OpenSCAP definitions file \
                                                       to do the evaluation with. \
                                                       Must use protocol file://|http://|https://. \
                                                       Must have file extension .xml|.bz2.
`oscap-profile`                | No        |         | OpenSCAP profile to evaluate.
`oscap-tailoring-uri`          | No        |         | URI to OpenSCAP tailoring file \
                                                       to do the evaluation with. \
                                                       Must use protocol file://|http://|https://. \
                                                       Must have file extension .xml|.bz2.
`oscap-fetch-remote-resources` | No        | True    | For Source DataStream and XCCDF files \
                                                       that have remote references fetch them if \
                                                       True, else don't. \
                                                       <br/><br/> \
                                                       **WARNING**: evaluations will not be \
                                                       complete if input defintions require \
                                                       remote resources and this is not True. \
                                                       For disconnected environments the remote \
                                                       internal mirror.
`[container-image-pull-repository-type, container-image-repository-type]` \
                               | Yes       | 'containers-storage:' \
                                                     | \
                                           Container repository type for the pull image source. \
                                           See https://github.com/containers/skopeo for valid \
                                           options.

Result Artifacts
----------------
Results artifacts output by this step.

Result Artifact Key | Description
--------------------|------------
`html-report`       | HTML report generated by oscap eval
`xml-report`        | XML report generated by oscap eval
`stdout-report`     | stdout report generated by oscap eval
"""

from ploigos_step_runner.step_implementers.shared.openscap_generic import OpenSCAPGeneric


class OpenSCAP(OpenSCAPGeneric):
    """`StepImplementer` for the `container-image-static-vulnerability-scan` step using OpenSCAP.

    Convenience wrapper for the `OpenSCAPGeneric` step implementer so that users
    can specify `OpenSCAP` for the `container-image-static-compliance-scan`.
    """




import os
from unittest.mock import PropertyMock, patch

from ploigos_step_runner import StepResult, WorkflowResult, StepRunnerException
from ploigos_step_runner.step_implementers.push_artifacts import MavenDeploy
from testfixtures import TempDirectory
from tests.helpers.base_step_implementer_test_case import \
    BaseStepImplementerTestCase


@patch("ploigos_step_runner.step_implementers.shared.MavenGeneric.__init__")
class TestStepImplementerMavenDeploy___init__(BaseStepImplementerTestCase):
    def test_defaults(self, mock_super_init):
        workflow_result = WorkflowResult()
        parent_work_dir_path = '/fake/path'
        config = {}

        MavenDeploy(
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path,
            config=config
        )

        mock_super_init.assert_called_once_with(
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path,
            config=config,
            environment=None,
            maven_phases_and_goals=['deploy']
        )

    def test_given_environment(self, mock_super_init):
        workflow_result = WorkflowResult()
        parent_work_dir_path = '/fake/path'
        config = {}

        MavenDeploy(
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path,
            config=config,
            environment='mock-env'
        )

        mock_super_init.assert_called_once_with(
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path,
            config=config,
            environment='mock-env',
            maven_phases_and_goals=['deploy']
        )

class TestStepImplementerMavenDeploy_step_implementer_config_defaults(
    BaseStepImplementerTestCase
):
    def test_result(self):
        self.assertEqual(
            MavenDeploy.step_implementer_config_defaults(),
            {
                'pom-file': 'pom.xml',
                'tls-verify': True,
                'maven-profiles': [],
                'maven-additional-arguments': [],
                'maven-no-transfer-progress': True,
                'maven-additional-arguments': [
                    '-Dmaven.install.skip=true',
                    '-Dmaven.test.skip=true'
                    '-DskipTests',
                    '-DskipITs'
                ]
            }
        )

class TestStepImplementerMavenDeploy__required_config_or_result_keys(
    BaseStepImplementerTestCase
):
    def test_result(self):
        self.assertEqual(
            MavenDeploy._required_config_or_result_keys(),
            [
                'pom-file',
                'maven-push-artifact-repo-url',
                'maven-push-artifact-repo-id',
                'version'
            ]
        )

@patch('ploigos_step_runner.step_implementers.push_artifacts.maven_deploy.run_maven')
@patch.object(MavenDeploy, '_run_maven_step')
@patch.object(
    MavenDeploy,
    'write_working_file',
    side_effect=['/mock/mvn_versions_set_output.txt', '/mock/mvn_deploy_output.txt']
)
@patch.object(
    MavenDeploy,
    'maven_settings_file',
    new_callable=PropertyMock,
    return_value='/fake/settings.xml'
)
class TestStepImplementerMavenDeploy__run_step(
    BaseStepImplementerTestCase
):
    def create_step_implementer(
            self,
            step_config={},
            workflow_result=None,
            parent_work_dir_path=''
    ):
        return self.create_given_step_implementer(
            step_implementer=MavenDeploy,
            step_config=step_config,
            step_name='deploy',
            implementer='MavenDeploy',
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path
        )

    def test_success(
        self,
        mock_settings_file,
        mock_write_working_file,
        mock_run_maven_step,
        mock_run_maven
    ):
        with TempDirectory() as test_dir:
            parent_work_dir_path = os.path.join(test_dir.path, 'working')

            pom_file = os.path.join(test_dir.path, 'mock-pom.xml')
            maven_push_artifact_repo_id = 'mock-repo-id'
            maven_push_artifact_repo_url = 'https://mock-repo.ploigos.com'
            version = '0.42.0-mock'
            step_config = {
                'pom-file': pom_file,
                'maven-push-artifact-repo-id': maven_push_artifact_repo_id,
                'maven-push-artifact-repo-url': maven_push_artifact_repo_url,
                'version': version
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            # run step
            actual_step_result = step_implementer._run_step()

            # create expected step result
            expected_step_result = StepResult(
                step_name='deploy',
                sub_step_name='MavenDeploy',
                sub_step_implementer_name='MavenDeploy'
            )
            expected_step_result.add_artifact(
                description="Standard out and standard error from running maven to update version.",
                name='maven-update-version-output',
                value='/mock/mvn_versions_set_output.txt'
            )
            expected_step_result.add_artifact(
                description="Standard out and standard error from running maven to " \
                    "push artifacts to repository.",
                name='maven-push-artifacts-output',
                value='/mock/mvn_deploy_output.txt'
            )

            # verify step result
            self.assertEqual(
                actual_step_result,
                expected_step_result
            )

            mock_write_working_file.assert_called()
            mock_run_maven.assert_called_with(
                mvn_output_file_path='/mock/mvn_versions_set_output.txt',
                settings_file='/fake/settings.xml',
                pom_file=pom_file,
                phases_and_goals=['versions:set'],
                additional_arguments=[
                    f'-DnewVersion={version}'
                ]
            )
            mock_run_maven_step.assert_called_with(
                mvn_output_file_path='/mock/mvn_deploy_output.txt',
                step_implementer_additional_arguments=[
                    '-DaltDeploymentRepository=' \
                    f'{maven_push_artifact_repo_id}::default::{maven_push_artifact_repo_url}'
                ]
            )

    def test_fail_set_version(
        self,
        mock_settings_file,
        mock_write_working_file,
        mock_run_maven_step,
        mock_run_maven
    ):
        with TempDirectory() as test_dir:
            parent_work_dir_path = os.path.join(test_dir.path, 'working')

            pom_file = os.path.join(test_dir.path, 'mock-pom.xml')
            maven_push_artifact_repo_id = 'mock-repo-id'
            maven_push_artifact_repo_url = 'https://mock-repo.ploigos.com'
            version = '0.42.0-mock'
            step_config = {
                'pom-file': pom_file,
                'maven-push-artifact-repo-id': maven_push_artifact_repo_id,
                'maven-push-artifact-repo-url': maven_push_artifact_repo_url,
                'version': version
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            # run step with mvn version:set failure
            mock_run_maven.side_effect = StepRunnerException('mock error setting new pom version')
            actual_step_result = step_implementer._run_step()

            # create expected step result
            expected_step_result = StepResult(
                step_name='deploy',
                sub_step_name='MavenDeploy',
                sub_step_implementer_name='MavenDeploy'
            )
            expected_step_result.success = False
            expected_step_result.message = "Error running 'maven deploy' to push artifacts. " \
                "More details maybe found in 'maven-output' report artifact: " \
                "mock error setting new pom version"
            expected_step_result.add_artifact(
                description="Standard out and standard error from running maven to update version.",
                name='maven-update-version-output',
                value='/mock/mvn_versions_set_output.txt'
            )
            expected_step_result.add_artifact(
                description="Standard out and standard error from running maven to " \
                    "push artifacts to repository.",
                name='maven-push-artifacts-output',
                value='/mock/mvn_deploy_output.txt'
            )

            # verify step result
            self.assertEqual(
                actual_step_result,
                expected_step_result
            )

            mock_write_working_file.assert_called()
            mock_run_maven.assert_called_with(
                mvn_output_file_path='/mock/mvn_versions_set_output.txt',
                settings_file='/fake/settings.xml',
                pom_file=pom_file,
                phases_and_goals=['versions:set'],
                additional_arguments=[
                    f'-DnewVersion={version}'
                ]
            )
            mock_run_maven_step.assert_not_called()

    def test_fail_mvn_depoy(
        self,
        mock_settings_file,
        mock_write_working_file,
        mock_run_maven_step,
        mock_run_maven
    ):
        with TempDirectory() as test_dir:
            parent_work_dir_path = os.path.join(test_dir.path, 'working')

            pom_file = os.path.join(test_dir.path, 'mock-pom.xml')
            maven_push_artifact_repo_id = 'mock-repo-id'
            maven_push_artifact_repo_url = 'https://mock-repo.ploigos.com'
            version = '0.42.0-mock'
            step_config = {
                'pom-file': pom_file,
                'maven-push-artifact-repo-id': maven_push_artifact_repo_id,
                'maven-push-artifact-repo-url': maven_push_artifact_repo_url,
                'version': version
            }
            step_implementer = self.create_step_implementer(
                step_config=step_config,
                parent_work_dir_path=parent_work_dir_path,
            )

            # run step with mvn deploy failure
            mock_run_maven_step.side_effect = StepRunnerException('mock error running mvn deploy')
            actual_step_result = step_implementer._run_step()

            # create expected step result
            expected_step_result = StepResult(
                step_name='deploy',
                sub_step_name='MavenDeploy',
                sub_step_implementer_name='MavenDeploy'
            )
            expected_step_result.success = False
            expected_step_result.message = "Error running 'maven deploy' to push artifacts. " \
                "More details maybe found in 'maven-output' report artifact: " \
                "mock error running mvn deploy"
            expected_step_result.add_artifact(
                description="Standard out and standard error from running maven to update version.",
                name='maven-update-version-output',
                value='/mock/mvn_versions_set_output.txt'
            )
            expected_step_result.add_artifact(
                description="Standard out and standard error from running maven to " \
                    "push artifacts to repository.",
                name='maven-push-artifacts-output',
                value='/mock/mvn_deploy_output.txt'
            )

            # verify step result
            self.assertEqual(
                actual_step_result,
                expected_step_result
            )

            mock_write_working_file.assert_called()
            mock_run_maven.assert_called_with(
                mvn_output_file_path='/mock/mvn_versions_set_output.txt',
                settings_file='/fake/settings.xml',
                pom_file=pom_file,
                phases_and_goals=['versions:set'],
                additional_arguments=[
                    f'-DnewVersion={version}'
                ]
            )
            mock_run_maven_step.assert_called_with(
                mvn_output_file_path='/mock/mvn_deploy_output.txt',
                step_implementer_additional_arguments=[
                    '-DaltDeploymentRepository=' \
                    f'{maven_push_artifact_repo_id}::default::{maven_push_artifact_repo_url}'
                ]
            )



from io import StringIO
import json
import os.path
import sh

import unittest
from testfixtures import TempDirectory
from unittest.mock import patch

from tests.helpers.base_test_case import BaseTestCase
from tests.helpers.sops_integration_test_case import SOPSIntegrationTestCase
from tests.helpers.test_utils import Any

from ploigos_step_runner.config.config_value import ConfigValue
from ploigos_step_runner.config.decryptors.sops import SOPS
from ploigos_step_runner.utils.file import parse_yaml_or_json_file

@patch('sh.sops', create=True)
class TestSOPSConfigValueDecryptor(BaseTestCase):
    def test_can_decrypt_true(self, sops_mock):
        encrypted_config_file_path = os.path.join(
            os.path.dirname(__file__),
            'files',
            'step-runner-config-secret-stuff.yml'
        )

        config_value = ConfigValue(
            value='ENC[AES256_GCM,data:UGKfnzsSrciR7GXZJhOCMmFrz3Y6V3pZsd3P,iv:yuReqA+n+rRXVHMc+2US5t7yPx54sooZSXWV4KLjDIs=,tag:jueP7/ZWLfYrEuhh+4eS8g==,type:str]',
            parent_source=encrypted_config_file_path,
            path_parts=['step-runner-config', 'global-environment-defaults', 'DEV', 'kube-api-token']
        )

        sops_decryptor = SOPS()
        self.assertTrue(
            sops_decryptor.can_decrypt(config_value)
        )

    def test_can_decrypt_false(self, sops_mock):
        encrypted_config_file_path = os.path.join(
            os.path.dirname(__file__),
            'files',
            'step-runner-config-secret-stuff.yml'
        )

        config_value = ConfigValue(
            value='not a sops encrypted value',
            parent_source=encrypted_config_file_path,
            path_parts=['step-runner-config', 'global-environment-defaults', 'DEV', 'kube-api-token']
        )

        sops_decryptor = SOPS()
        self.assertFalse(
            sops_decryptor.can_decrypt(config_value)
        )

    def test_can_can_decrypt_not_string(self, sops_mock):
        encrypted_config_file_path = os.path.join(
            os.path.dirname(__file__),
            'files',
            'step-runner-config-secret-stuff.yml'
        )

        config_value = ConfigValue(
            value=True,
            parent_source=encrypted_config_file_path,
            path_parts=['step-runner-config', 'global-environment-defaults', 'DEV', 'kube-api-token']
        )

        sops_decryptor = SOPS()
        self.assertFalse(
            sops_decryptor.can_decrypt(config_value)
        )

    def test_decrypt_parent_source_file(self, sops_mock):
        encrypted_config_file_path = os.path.join(
            os.path.dirname(__file__),
            'files',
            'step-runner-config-secret-stuff.yml'
        )

        config_value = ConfigValue(
            value='ENC[AES256_GCM,data:UGKfnzsSrciR7GXZJhOCMmFrz3Y6V3pZsd3P,iv:yuReqA+n+rRXVHMc+2US5t7yPx54sooZSXWV4KLjDIs=,tag:jueP7/ZWLfYrEuhh+4eS8g==,type:str]',
            parent_source=encrypted_config_file_path,
            path_parts=['step-runner-config', 'global-environment-defaults', 'DEV', 'kube-api-token']
        )

        sops_decryptor = SOPS()

        sops_decryptor.decrypt(config_value)
        sops_mock.assert_called_once_with(
            '--decrypt',
            '--extract=["step-runner-config"]["global-environment-defaults"]["DEV"]["kube-api-token"]',
            None,
            encrypted_config_file_path,
            _in=None,
            _out=Any(StringIO),
            _err=Any(StringIO)
        )

    def test_decrypt_additional_sops_args(self, sops_mock):
        encrypted_config_file_path = os.path.join(
            os.path.dirname(__file__),
            'files',
            'step-runner-config-secret-stuff.yml'
        )

        config_value = ConfigValue(
            value='ENC[AES256_GCM,data:UGKfnzsSrciR7GXZJhOCMmFrz3Y6V3pZsd3P,iv:yuReqA+n+rRXVHMc+2US5t7yPx54sooZSXWV4KLjDIs=,tag:jueP7/ZWLfYrEuhh+4eS8g==,type:str]',
            parent_source=encrypted_config_file_path,
            path_parts=['step-runner-config', 'global-environment-defaults', 'DEV', 'kube-api-token']
        )

        sops_decryptor = SOPS(
            additional_sops_args=[
                '--aws-profile=foo'
            ]
        )

        sops_decryptor.decrypt(config_value)
        sops_mock.assert_called_once_with(
            '--decrypt',
            '--extract=["step-runner-config"]["global-environment-defaults"]["DEV"]["kube-api-token"]',
            None,
            encrypted_config_file_path,
            '--aws-profile=foo',
            _in=None,
            _out=Any(StringIO),
            _err=Any(StringIO)
        )

    def test_decrypt_parent_source_dict(self, sops_mock):
        encrypted_config_file_path = os.path.join(
            os.path.dirname(__file__),
            'files',
            'step-runner-config-secret-stuff.yml'
        )

        encrypted_config = parse_yaml_or_json_file(encrypted_config_file_path)
        encrypted_config_json = json.dumps(encrypted_config)

        config_value = ConfigValue(
            value='ENC[AES256_GCM,data:UGKfnzsSrciR7GXZJhOCMmFrz3Y6V3pZsd3P,iv:yuReqA+n+rRXVHMc+2US5t7yPx54sooZSXWV4KLjDIs=,tag:jueP7/ZWLfYrEuhh+4eS8g==,type:str]',
            parent_source=encrypted_config,
            path_parts=['step-runner-config', 'global-environment-defaults', 'DEV', 'kube-api-token']
        )

        sops_decryptor = SOPS()

        sops_decryptor.decrypt(config_value)
        sops_mock.assert_called_once_with(
            '--decrypt',
            '--extract=["step-runner-config"]["global-environment-defaults"]["DEV"]["kube-api-token"]',
            '--input-type=json',
            '/dev/stdin',
            _in=encrypted_config_json,
            _out=Any(StringIO),
            _err=Any(StringIO)
        )

    def test_decrypt_parent_source_file_does_not_exist(self, sops_mock):
        config_value = ConfigValue(
            value='ENC[AES256_GCM,data:UGKfnzsSrciR7GXZJhOCMmFrz3Y6V3pZsd3P,iv:yuReqA+n+rRXVHMc+2US5t7yPx54sooZSXWV4KLjDIs=,tag:jueP7/ZWLfYrEuhh+4eS8g==,type:str]',
            parent_source='does-not-exist.yml',
            path_parts=['step-runner-config', 'global-environment-defaults', 'DEV', 'kube-api-token']
        )

        sops_decryptor = SOPS()

        with self.assertRaisesRegex(
            ValueError,
            r"Given config value \(ConfigValue\(.*\)\) parent source \(does-not-exist.yml\)" \
            r" is of type \(str\) but is not a path to a file that exists"
        ):
            sops_decryptor.decrypt(config_value)

    def test_decrypt_parent_source_none(self, sops_mock):
        config_value = ConfigValue(
            value='ENC[AES256_GCM,data:UGKfnzsSrciR7GXZJhOCMmFrz3Y6V3pZsd3P,iv:yuReqA+n+rRXVHMc+2US5t7yPx54sooZSXWV4KLjDIs=,tag:jueP7/ZWLfYrEuhh+4eS8g==,type:str]',
            parent_source=None,
            path_parts=['step-runner-config', 'global-environment-defaults', 'DEV', 'kube-api-token']
        )

        sops_decryptor = SOPS()

        with self.assertRaisesRegex(
            ValueError,
            r"Given config value \(ConfigValue\(.*\)\) parent source \(None\) " \
            r"is expected to be of type dict or str but is of type: <class 'NoneType'>"
        ):
            sops_decryptor.decrypt(config_value)

    def test_decrypt_sops_error(self, sops_mock):
        encrypted_config_file_path = os.path.join(
            os.path.dirname(__file__),
            'files',
            'step-runner-config-secret-stuff.yml'
        )

        config_value = ConfigValue(
            value='ENC[AES256_GCM,data:UGKfnzsSrciR7GXZJhOCMmFrz3Y6V3pZsd3P,iv:yuReqA+n+rRXVHMc+2US5t7yPx54sooZSXWV4KLjDIs=,tag:jueP7/ZWLfYrEuhh+4eS8g==,type:str]',
            parent_source=encrypted_config_file_path,
            path_parts=['step-runner-config', 'global-environment-defaults', 'DEV', 'kube-api-token']
        )

        sops_decryptor = SOPS()

        sh.sops.side_effect = sh.ErrorReturnCode('sops', b'mock stdout', b'mock error about issue running sops')
        with self.assertRaisesRegex(
            RuntimeError,
            r"Error invoking sops when trying to decrypt config value \(ConfigValue\(.*\)\):"
        ):
            sops_decryptor.decrypt(config_value)

    def test_get_sops_value_path(self, sops_mock):
        config_value = ConfigValue(
            value='ENC[AES256_GCM,data:UGKfnzsSrciR7GXZJhOCMmFrz3Y6V3pZsd3P,iv:yuReqA+n+rRXVHMc+2US5t7yPx54sooZSXWV4KLjDIs=,tag:jueP7/ZWLfYrEuhh+4eS8g==,type:str]',
            parent_source=None,
            path_parts=["step-runner-config", "step-foo", 0, "config", "test1"]
        )

        sops_value_path = SOPS.get_sops_value_path(config_value)
        self.assertEqual(
            sops_value_path,
            '["step-runner-config"]["step-foo"][0]["config"]["test1"]')

class TestSOPSConfigValueDecryptorSOPSIntegrationTests(SOPSIntegrationTestCase):
    def test_can_decrypt_true(self):
        encrypted_config_file_path = os.path.join(
            os.path.dirname(__file__),
            'files',
            'step-runner-config-secret-stuff.yml'
        )

        config_value = ConfigValue(
            value='ENC[AES256_GCM,data:UGKfnzsSrciR7GXZJhOCMmFrz3Y6V3pZsd3P,iv:yuReqA+n+rRXVHMc+2US5t7yPx54sooZSXWV4KLjDIs=,tag:jueP7/ZWLfYrEuhh+4eS8g==,type:str]',
            parent_source=encrypted_config_file_path,
            path_parts=['step-runner-config', 'global-environment-defaults', 'DEV', 'kube-api-token']
        )

        sops_decryptor = SOPS()
        self.assertTrue(
            sops_decryptor.can_decrypt(config_value)
        )

    def test_can_decrypt_false(self):
        encrypted_config_file_path = os.path.join(
            os.path.dirname(__file__),
            'files',
            'step-runner-config-secret-stuff.yml'
        )

        config_value = ConfigValue(
            value='not a sops encrypted value',
            parent_source=encrypted_config_file_path,
            path_parts=['step-runner-config', 'global-environment-defaults', 'DEV', 'kube-api-token']
        )

        sops_decryptor = SOPS()
        self.assertFalse(
            sops_decryptor.can_decrypt(config_value)
        )

    def test_decrypt_parent_source_file(self):
        encrypted_config_file_path = os.path.join(
            os.path.dirname(__file__),
            'files',
            'step-runner-config-secret-stuff.yml'
        )

        config_value = ConfigValue(
            value='ENC[AES256_GCM,data:UGKfnzsSrciR7GXZJhOCMmFrz3Y6V3pZsd3P,iv:yuReqA+n+rRXVHMc+2US5t7yPx54sooZSXWV4KLjDIs=,tag:jueP7/ZWLfYrEuhh+4eS8g==,type:str]',
            parent_source=encrypted_config_file_path,
            path_parts=['step-runner-config', 'global-environment-defaults', 'DEV', 'kube-api-token']
        )

        sops_decryptor = SOPS()

        decrypted_value = sops_decryptor.decrypt(config_value)

        self.assertEqual(
            decrypted_value,
            'super secret kube api token'
        )

    def test_decrypt_parent_source_dict(self):
        encrypted_config_file_path = os.path.join(
            os.path.dirname(__file__),
            'files',
            'step-runner-config-secret-stuff.yml'
        )

        encrypted_config = parse_yaml_or_json_file(encrypted_config_file_path)

        config_value = ConfigValue(
            value='ENC[AES256_GCM,data:UGKfnzsSrciR7GXZJhOCMmFrz3Y6V3pZsd3P,iv:yuReqA+n+rRXVHMc+2US5t7yPx54sooZSXWV4KLjDIs=,tag:jueP7/ZWLfYrEuhh+4eS8g==,type:str]',
            parent_source=encrypted_config,
            path_parts=['step-runner-config', 'global-environment-defaults', 'DEV', 'kube-api-token']
        )

        sops_decryptor = SOPS()

        decrypted_value = sops_decryptor.decrypt(config_value)

        self.assertEqual(
            decrypted_value,
            'super secret kube api token'
        )

    def test_decrypt_parent_source_file_does_not_exist(self):
        config_value = ConfigValue(
            value='ENC[AES256_GCM,data:UGKfnzsSrciR7GXZJhOCMmFrz3Y6V3pZsd3P,iv:yuReqA+n+rRXVHMc+2US5t7yPx54sooZSXWV4KLjDIs=,tag:jueP7/ZWLfYrEuhh+4eS8g==,type:str]',
            parent_source='does-not-exist.yml',
            path_parts=['step-runner-config', 'global-environment-defaults', 'DEV', 'kube-api-token']
        )

        sops_decryptor = SOPS()

        with self.assertRaisesRegex(
            ValueError,
            r"Given config value \(ConfigValue\(.*\)\) parent source \(does-not-exist.yml\)" \
            r" is of type \(str\) but is not a path to a file that exists"
        ):
            sops_decryptor.decrypt(config_value)

    def test_decrypt_parent_source_none(self):
        config_value = ConfigValue(
            value='ENC[AES256_GCM,data:UGKfnzsSrciR7GXZJhOCMmFrz3Y6V3pZsd3P,iv:yuReqA+n+rRXVHMc+2US5t7yPx54sooZSXWV4KLjDIs=,tag:jueP7/ZWLfYrEuhh+4eS8g==,type:str]',
            parent_source=None,
            path_parts=['step-runner-config', 'global-environment-defaults', 'DEV', 'kube-api-token']
        )

        sops_decryptor = SOPS()

        with self.assertRaisesRegex(
            ValueError,
            r"Given config value \(ConfigValue\(.*\)\) parent source \(None\) " \
            r"is expected to be of type dict or str but is of type: <class 'NoneType'>"
        ):
            sops_decryptor.decrypt(config_value)

    def test_decrypt_no_valid_key(self):
        encrypted_config_file_path = os.path.join(
            os.path.dirname(__file__),
            'files',
            'step-runner-config-secret-stuff.yml'
        )

        config_value = ConfigValue(
            value='ENC[AES256_GCM,data:UGKfnzsSrciR7GXZJhOCMmFrz3Y6V3pZsd3P,iv:yuReqA+n+rRXVHMc+2US5t7yPx54sooZSXWV4KLjDIs=,tag:jueP7/ZWLfYrEuhh+4eS8g==,type:str]',
            parent_source=encrypted_config_file_path,
            path_parts=['step-runner-config', 'global-environment-defaults', 'DEV', 'kube-api-token']
        )

        sops_decryptor = SOPS()

        # delete the gpg key needed to decrypt the value
        self.delete_gpg_key()

        # attempt to decrypt the value
        with self.assertRaisesRegex(
            RuntimeError,
            r"Error invoking sops when trying to decrypt config value \(ConfigValue\(.*\)\):"
        ):
            sops_decryptor.decrypt(config_value)


"""`StepImplementer` for the `uat` step using Maven by invoking the 'test` maven phase."

Step Configuration
------------------
Step configuration expected as input to this step.
Could come from:
* static configuration
* runtime configuration
* previous step results

Configuration Key            | Required? | Default | Description
-----------------------------|-----------|---------|-----------
`pom-file`                   | Yes       | `'pom.xml'` | pom used when executing maven.
`tls-verify`                 | No        | `True`  | Disables TLS Verification if set to False
`maven-profiles`             | No        | `[]`    | List of maven profiles to use.
`maven-no-transfer-progress` | No        | `True`  | `True` to suppress the transfer progress of packages maven downloads.
                                                     `False` to have the transfer progress printed.\
                                                      See https://maven.apache.org/ref/current/maven-embedder/cli.html
`maven-additional-arguments` | No        | `[]`    | List of additional arguments to use.
`maven-servers`              | No        |         | Dictionary of dictionaries of id, username, password
`maven-repositories`         | No        |         | Dictionary of dictionaries of id, url, snapshots, releases
`maven-mirrors`              | No        |         | Dictionary of dictionaries of id, url, mirror_of
`test-reports-dir`           | No        |         | Default is to try and dynamically determine where the test reports directory is \
                                                     based on configuration in the given `pom-file`, but it is impossible task to do \
                                                     in all cases. \
                                                     So, this parameter provides a way for the user to specify where the test results \
                                                     are if our attempts at dynamically figuring it out are failing your unique pom.


Result Artifacts
----------------
Results artifacts output by this step.

Result Artifact Key | Description
--------------------|------------
`maven-output`      | Path to Stdout and Stderr from invoking Maven.
`test-report`       | Directory containing the test reports generated from running this step.
"""# pylint: disable=line-too-long

from ploigos_step_runner import StepResult, StepRunnerException
from ploigos_step_runner.step_implementers.shared import (
    MavenGeneric, MavenTestReportingMixin)

DEFAULT_CONFIG = {}

REQUIRED_CONFIG_OR_PREVIOUS_STEP_RESULT_ARTIFACT_KEYS = [
    'pom-file'
]

class MavenTest(MavenGeneric, MavenTestReportingMixin):
    """`StepImplementer` for the `uat` step using Maven by invoking the 'test` maven phase.
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        workflow_result,
        parent_work_dir_path,
        config,
        environment=None
    ):
        super().__init__(
            workflow_result=workflow_result,
            parent_work_dir_path=parent_work_dir_path,
            config=config,
            environment=environment,
            maven_phases_and_goals=['test']
        )

    @staticmethod
    def step_implementer_config_defaults():
        """Getter for the StepImplementer's configuration defaults.

        Returns
        -------
        dict
            Default values to use for step configuration values.

        Notes
        -----
        These are the lowest precedence configuration values.
        """
        return {**MavenGeneric.step_implementer_config_defaults(), **DEFAULT_CONFIG}

    @staticmethod
    def _required_config_or_result_keys():
        """Getter for step configuration or previous step result artifacts that are required before
        running this step.

        See Also
        --------
        _validate_required_config_or_previous_step_result_artifact_keys

        Returns
        -------
        array_list
            Array of configuration keys or previous step result artifacts
            that are required before running the step.
        """
        return REQUIRED_CONFIG_OR_PREVIOUS_STEP_RESULT_ARTIFACT_KEYS

    def _run_step(self):
        """Runs the step implemented by this StepImplementer.

        Returns
        -------
        StepResult
            Object containing the dictionary results of this step.
        """
        step_result = StepResult.from_step_implementer(self)

        # run the tests
        print("Run unit tests")
        mvn_output_file_path = self.write_working_file('mvn_output.txt')
        try:
            # execute maven step (params come from config)
            self._run_maven_step(
                mvn_output_file_path=mvn_output_file_path
            )
        except StepRunnerException as error:
            step_result.success = False
            step_result.message = "Error running maven. " \
                f"More details maybe found in report artifacts: {error}"
        finally:
            step_result.add_artifact(
                description="Standard out and standard error from maven.",
                name='maven-output',
                value=mvn_output_file_path
            )

        # get test report dir
        test_report_dir = self.__get_test_report_dir()
        if test_report_dir:
            step_result.add_artifact(
                description="Test report generated when running unit tests.",
                name='test-report',
                value=test_report_dir
            )

            # gather test report evidence
            self._gather_evidence_from_test_report_directory_testsuite_elements(
                step_result=step_result,
                test_report_dir=test_report_dir
            )

        # return result
        return step_result

    def __get_test_report_dir(self):
        """Gets the test report directory.

        Search Priority:
        * values -> 'test-reports-dir'
        * pom.xml -> maven-surefire-plugin -> reportsDirectory

        Returns
        -------
        str
            Path to the directory containing the test reports.
        """
        # user supplied where the test reports go, just use that
        test_report_dir = self.get_value('test-reports-dir')

        # else do our best to find them
        if not test_report_dir:
            # attempt to get failsafe test report dir, if not, try for surefire
            test_report_dir = None
            try:
                test_report_dir = self._attempt_get_test_report_directory(
                    plugin_name=MavenTestReportingMixin.SUREFIRE_PLUGIN_NAME,
                    configuration_key=\
                        MavenTestReportingMixin.SUREFIRE_PLUGIN_REPORTS_DIR_CONFIG_NAME,
                    default=MavenTestReportingMixin.SUREFIRE_PLUGIN_DEFAULT_REPORTS_DIR
                )
            except StepRunnerException:
                print(
                    'WARNING: Did not find any expected test reporting plugin'
                    f' ({MavenTestReportingMixin.SUREFIRE_PLUGIN_NAME})'
                    ' to read artifacts and evidence from.'
                    ' This is not wholly unexpected because there is enumerable maven plugins,'
                    ' and enumerable ways to configure them.'
                    ' Rather then relying on this step implementer to try and figure out'
                    ' where the test reports are you can configure it manually via the'
                    ' step implementer config (test-reports-dir).'
                )

        return test_report_dir



import re
from io import IOBase
from unittest.mock import call, patch

import sh
from ploigos_step_runner.config import ConfigValue
from ploigos_step_runner.utils.containers import *
from tests.helpers.base_test_case import BaseTestCase
from tests.helpers.test_utils import *


def create_which_side_effect(cmd, cmd_path):
    def which_side_effect(*args, **kwargs):
        if args[0] == cmd:
            return cmd_path
        else:
            return None

    return which_side_effect

class TestContainerRegistryLogin(BaseTestCase):
    @patch('sh.buildah', create=True)
    @patch('sh.which', create=True)
    def test_buildah(self, which_mock, container_command_mock):
        which_mock.side_effect = create_which_side_effect(
            cmd='buildah',
            cmd_path='/mock/buildah'
        )

        container_registry_login(
            container_registry_uri='registry.example.xyz',
            container_registry_username='example',
            container_registry_password='nope'
        )

        container_command_mock.bake.assert_called_once()
        container_command_mock.bake().login.bake.assert_called_once_with(
            password_stdin=True,
            username='example',
            tls_verify='true'
        )
        container_command_mock.bake().login.bake().assert_called_once_with(
            'registry.example.xyz',
            _in='nope',
            _out=Any(IOBase),
            _err=Any(IOBase),
            _tee='err'
        )

    @patch('sh.podman', create=True)
    @patch('sh.which', create=True)
    def test_podman(self, which_mock, container_command_mock):
        which_mock.side_effect = create_which_side_effect(
            cmd='podman',
            cmd_path='/mock/podman'
        )

        container_registry_login(
            container_registry_uri='registry.example.xyz',
            container_registry_username='example',
            container_registry_password='nope'
        )

        container_command_mock.bake.assert_called_once()
        container_command_mock.bake().login.bake.assert_called_once_with(
            password_stdin=True,
            username='example',
            tls_verify='true'
        )
        container_command_mock.bake().login.bake().assert_called_once_with(
            'registry.example.xyz',
            _in='nope',
            _out=Any(IOBase),
            _err=Any(IOBase),
            _tee='err'
        )

    @patch('sh.skopeo', create=True)
    @patch('sh.which', create=True)
    def test_skopeo(self, which_mock, container_command_mock):
        which_mock.side_effect = create_which_side_effect(
            cmd='skopeo',
            cmd_path='/mock/skopeo'
        )

        container_registry_login(
            container_registry_uri='registry.example.xyz',
            container_registry_username='example',
            container_registry_password='nope'
        )

        container_command_mock.bake.assert_called_once()
        container_command_mock.bake().login.bake.assert_called_once_with(
            password_stdin=True,
            username='example',
            tls_verify='true'
        )
        container_command_mock.bake().login.bake().assert_called_once_with(
            'registry.example.xyz',
            _in='nope',
            _out=Any(IOBase),
            _err=Any(IOBase),
            _tee='err'
        )

    @patch('sh.Command')
    @patch('sh.which', create=True)
    def test_use_given_command_exists(self, which_mock, container_command_mock):
        which_mock.side_effect = create_which_side_effect(
            cmd='fake-podman',
            cmd_path='/mock/fake-podman'
        )

        container_registry_login(
            container_registry_uri='registry.example.xyz',
            container_registry_username='example',
            container_registry_password='nope',
            container_command_short_name='fake-podman'
        )

        container_command_mock().bake.assert_called_once()
        container_command_mock().bake().login.bake.assert_called_once_with(
            password_stdin=True,
            username='example',
            tls_verify='true'
        )
        container_command_mock().bake().login.bake().assert_called_once_with(
            'registry.example.xyz',
            _in='nope',
            _out=Any(IOBase),
            _err=Any(IOBase),
            _tee='err'
        )

    @patch('sh.Command')
    @patch('sh.which', create=True)
    def test_use_given_command_does_not_exists(self, which_mock, container_command_mock):
        which_mock.side_effect = create_which_side_effect(
            cmd=None,
            cmd_path=None
        )

        with self.assertRaisesRegex(
            RuntimeError,
            r"When attempting to login to container registry \(registry.example.xyz\) "
            r"could not find the given expected tool \(fake-podman\) to login with"
        ):
            container_registry_login(
                container_registry_uri='registry.example.xyz',
                container_registry_username='example',
                container_registry_password='nope',
                container_command_short_name='fake-podman'
            )

    @patch('sh.buildah', create=True)
    @patch('sh.which', create=True)
    def test_config_auth_param(self, which_mock, container_command_mock):
        which_mock.side_effect = create_which_side_effect(
            cmd='buildah',
            cmd_path='/mock/buildah'
        )

        container_registry_login(
            container_registry_uri='registry.example.xyz',
            container_registry_username='example',
            container_registry_password='nope',
            containers_config_auth_file='/tmp/test/auth.json'
        )

        container_command_mock.bake.assert_called_once()
        container_command_mock.bake().login.bake.assert_called_once_with(
            password_stdin=True,
            username='example',
            tls_verify='true',
            authfile='/tmp/test/auth.json'
        )
        container_command_mock.bake().login.bake().assert_called_once_with(
            'registry.example.xyz',
            _in='nope',
            _out=Any(IOBase),
            _err=Any(IOBase),
            _tee='err'
        )

    @patch('sh.buildah', create=True)
    @patch('sh.which', create=True)
    def test_tls_verify_param_bool_false(self, which_mock, container_command_mock):
        which_mock.side_effect = create_which_side_effect(
            cmd='buildah',
            cmd_path='/mock/buildah'
        )

        container_registry_login(
            container_registry_uri='registry.example.xyz',
            container_registry_username='example',
            container_registry_password='nope',
            container_registry_tls_verify=False
        )

        container_command_mock.bake.assert_called_once()
        container_command_mock.bake().login.bake.assert_called_once_with(
            password_stdin=True,
            username='example',
            tls_verify='false',
        )
        container_command_mock.bake().login.bake().assert_called_once_with(
            'registry.example.xyz',
            _in='nope',
            _out=Any(IOBase),
            _err=Any(IOBase),
            _tee='err'
        )

    @patch('sh.buildah', create=True)
    @patch('sh.which', create=True)
    def test_tls_verify_param_str_false(self, which_mock, container_command_mock):
        which_mock.side_effect = create_which_side_effect(
            cmd='buildah',
            cmd_path='/mock/buildah'
        )

        container_registry_login(
            container_registry_uri='registry.example.xyz',
            container_registry_username='example',
            container_registry_password='nope',
            container_registry_tls_verify="false"
        )

        container_command_mock.bake.assert_called_once()
        container_command_mock.bake().login.bake.assert_called_once_with(
            password_stdin=True,
            username='example',
            tls_verify='false',
        )
        container_command_mock.bake().login.bake().assert_called_once_with(
            'registry.example.xyz',
            _in='nope',
            _out=Any(IOBase),
            _err=Any(IOBase),
            _tee='err'
        )

    @patch('sh.buildah', create=True)
    @patch('sh.which', create=True)
    def test_tls_verify_param_bool_true(self, which_mock, container_command_mock):
        which_mock.side_effect = create_which_side_effect(
            cmd='buildah',
            cmd_path='/mock/buildah'
        )

        container_registry_login(
            container_registry_uri='registry.example.xyz',
            container_registry_username='example',
            container_registry_password='nope',
            container_registry_tls_verify=True
        )

        container_command_mock.bake.assert_called_once()
        container_command_mock.bake().login.bake.assert_called_once_with(
            password_stdin=True,
            username='example',
            tls_verify='true',
        )
        container_command_mock.bake().login.bake().assert_called_once_with(
            'registry.example.xyz',
            _in='nope',
            _out=Any(IOBase),
            _err=Any(IOBase),
            _tee='err'
        )

    @patch('sh.buildah', create=True)
    @patch('sh.which', create=True)
    def test_tls_verify_param_str_true(self, which_mock, container_command_mock):
        which_mock.side_effect = create_which_side_effect(
            cmd='buildah',
            cmd_path='/mock/buildah'
        )

        container_registry_login(
            container_registry_uri='registry.example.xyz',
            container_registry_username='example',
            container_registry_password='nope',
            container_registry_tls_verify="true"
        )

        container_command_mock.bake.assert_called_once()
        container_command_mock.bake().login.bake.assert_called_once_with(
            password_stdin=True,
            username='example',
            tls_verify='true',
        )
        container_command_mock.bake().login.bake().assert_called_once_with(
            'registry.example.xyz',
            _in='nope',
            _out=Any(IOBase),
            _err=Any(IOBase),
            _tee='err'
        )

    @patch('sh.buildah', create=True)
    @patch('sh.which', create=True)
    def test_configvalue_params(self, which_mock, container_command_mock):
        which_mock.side_effect = create_which_side_effect(
            cmd='buildah',
            cmd_path='/mock/buildah'
        )

        container_registry_login(
            container_registry_uri=ConfigValue('registry.example.xyz'),
            container_registry_username=ConfigValue('example'),
            container_registry_password=ConfigValue('nope'),
            container_registry_tls_verify=ConfigValue(True),
            containers_config_auth_file=ConfigValue('/tmp/test/auth.json')
        )

        container_command_mock.bake.assert_called_once()
        container_command_mock.bake().login.bake.assert_called_once_with(
            password_stdin=True,
            username='example',
            tls_verify='true',
            authfile='/tmp/test/auth.json'
        )
        container_command_mock.bake().login.bake().assert_called_once_with(
            'registry.example.xyz',
            _in='nope',
            _out=Any(IOBase),
            _err=Any(IOBase),
            _tee='err'
        )

    @patch('sh.which', create=True)
    def test_no_container_registry_login_tools_found(self, which_mock):
        # ensure none of the tools are found
        which_mock.side_effect = create_which_side_effect(
            cmd=None,
            cmd_path=None
        )

        with self.assertRaisesRegex(
            RuntimeError,
            r"When attempting to login to container registry \(registry.example.xyz\) "
            r"could not find one of the expected tools \(buildah, podman, skopeo\) to login with."
        ):
            container_registry_login(
                container_registry_uri='registry.example.xyz',
                container_registry_username='example',
                container_registry_password='nope'
            )

    @patch('sh.buildah', create=True)
    @patch('sh.which', create=True)
    def test_buildah_login_fail(self, which_mock, container_command_mock):
        which_mock.side_effect = create_which_side_effect(
            cmd='buildah',
            cmd_path='/mock/buildah'
        )

        with self.assertRaisesRegex(
            RuntimeError,
            re.compile(
                r"Failed to login to container registry \(registry.example.xyz\)"
                r" with username \(example\):"
                r".*RAN: buildah login"
                r".*STDOUT:"
                r".*mock stdout"
                r".*STDERR:"
                r".*mock stderr login error",
                re.DOTALL
            )
        ):
            container_command_mock.bake().login.bake().side_effect = sh.ErrorReturnCode(
                'buildah login',
                b'mock stdout',
                b'mock stderr login error'
            )
            container_registry_login(
                container_registry_uri='registry.example.xyz',
                container_registry_username='example',
                container_registry_password='nope'
            )

class TestContainerRegistriesLogin(BaseTestCase):
    @patch('ploigos_step_runner.utils.containers.container_registry_login')
    def test_dict_of_dicts(self, container_registry_login_mock):
        registries = {
            'registry.redhat.io': {
                'username': 'hello1@world.xyz',
                'password': 'nope1'
            },
            'registry.internal.example.xyz': {
                'username': 'hello2@example.xyz',
                'password': 'nope2'
            }
        }

        container_registries_login(registries)

        calls = [
            call(
                container_registry_uri='registry.redhat.io',
                container_registry_username='hello1@world.xyz',
                container_registry_password='nope1',
                container_registry_tls_verify=True,
                containers_config_auth_file=None,
                container_command_short_name=None
            ),
            call(
                container_registry_uri='registry.internal.example.xyz',
                container_registry_username='hello2@example.xyz',
                container_registry_password='nope2',
                container_registry_tls_verify=True,
                containers_config_auth_file=None,
                container_command_short_name=None
            )
        ]
        container_registry_login_mock.assert_has_calls(calls)

    @patch('ploigos_step_runner.utils.containers.container_registry_login')
    def test_dict_of_dicts_with_tls_verify_value(self, container_registry_login_mock):
        registries = {
            'registry.redhat.io': {
                'username': 'hello1@world.xyz',
                'password': 'nope1',
                'tls-verify': False
            },
            'registry.internal.example.xyz': {
                'username': 'hello2@example.xyz',
                'password': 'nope2',
                'tls-verify': True
            }
        }

        container_registries_login(registries)

        calls = [
            call(
                container_registry_uri='registry.redhat.io',
                container_registry_username='hello1@world.xyz',
                container_registry_password='nope1',
                container_registry_tls_verify=False,
                containers_config_auth_file=None,
                container_command_short_name=None
            ),
            call(
                container_registry_uri='registry.internal.example.xyz',
                container_registry_username='hello2@example.xyz',
                container_registry_password='nope2',
                container_registry_tls_verify=True,
                containers_config_auth_file=None,
                container_command_short_name=None
            )
        ]
        container_registry_login_mock.assert_has_calls(calls)

    @patch('ploigos_step_runner.utils.containers.container_registry_login')
    def test_list_of_dicts_with_containers_config_auth_file(self, container_registry_login_mock):
        registries = {
            'registry.redhat.io': {
                'username': 'hello1@world.xyz',
                'password': 'nope1',
                'tls-verify': False
            },
            'registry.internal.example.xyz': {
                'username': 'hello2@example.xyz',
                'password': 'nope2',
                'tls-verify': True
            }
        }

        container_registries_login(registries, '/tmp/mock/auth.json')

        calls = [
            call(
                container_registry_uri='registry.redhat.io',
                container_registry_username='hello1@world.xyz',
                container_registry_password='nope1',
                container_registry_tls_verify=True,
                containers_config_auth_file='/tmp/mock/auth.json'
            ),
            call(
                container_registry_uri='registry.internal.example.xyz',
                container_registry_username='hello2@example.xyz',
                container_registry_password='nope2',
                container_registry_tls_verify=True,
                containers_config_auth_file='/tmp/mock/auth.json'
            )
        ]
        container_registry_login_mock.assert_has_calls(calls)

    @patch('ploigos_step_runner.utils.containers.container_registry_login')
    def test_dict_of_config_values(self, container_registry_login_mock):
        registries = {
            'registry.redhat.io': ConfigValue({
                'username': 'hello1@world.xyz',
                'password': 'nope1'
            }),
            'registry.internal.example.xyz': ConfigValue({
                'username': 'hello2@example.xyz',
                'password': 'nope2'
            })
        }

        container_registries_login(registries)

        calls = [
            call(
                container_registry_uri='registry.redhat.io',
                container_registry_username='hello1@world.xyz',
                container_registry_password='nope1',
                container_registry_tls_verify=True,
                containers_config_auth_file=None,
                container_command_short_name=None,
            ),
            call(
                container_registry_uri='registry.internal.example.xyz',
                container_registry_username='hello2@example.xyz',
                container_registry_password='nope2',
                container_registry_tls_verify=True,
                containers_config_auth_file=None,
                container_command_short_name=None
            )
        ]
        container_registry_login_mock.assert_has_calls(calls)

    @patch('ploigos_step_runner.utils.containers.container_registry_login')
    def test_dict_of_dicts_with_uri_keys(self, container_registry_login_mock):
        registries = {
            'redhat': {
                'uri': 'registry.redhat.io',
                'username': 'hello1@world.xyz',
                'password': 'nope1'
            },
            'internal': {
                'uri': 'registry.internal.example.xyz',
                'username': 'hello2@example.xyz',
                'password': 'nope2'
            }
        }

        container_registries_login(registries)

        calls = [
            call(
                container_registry_uri='registry.redhat.io',
                container_registry_username='hello1@world.xyz',
                container_registry_password='nope1',
                container_registry_tls_verify=True,
                containers_config_auth_file=None,
                container_command_short_name=None
            ),
            call(
                container_registry_uri='registry.internal.example.xyz',
                container_registry_username='hello2@example.xyz',
                container_registry_password='nope2',
                container_registry_tls_verify=True,
                containers_config_auth_file=None,
                container_command_short_name=None
            )
        ]
        container_registry_login_mock.assert_has_calls(calls)

    @patch('ploigos_step_runner.utils.containers.container_registry_login')
    def test_list_of_dicts(self, container_registry_login_mock):
        registries = [
            {
                'uri': 'registry.redhat.io',
                'username': 'hello1@world.xyz',
                'password': 'nope1'
            },
            {
                'uri': 'registry.internal.example.xyz',
                'username': 'hello2@example.xyz',
                'password': 'nope2'
            }
        ]

        container_registries_login(registries)

        calls = [
            call(
                container_registry_uri='registry.redhat.io',
                container_registry_username='hello1@world.xyz',
                container_registry_password='nope1',
                container_registry_tls_verify=True,
                containers_config_auth_file=None,
                container_command_short_name=None
            ),
            call(
                container_registry_uri='registry.internal.example.xyz',
                container_registry_username='hello2@example.xyz',
                container_registry_password='nope2',
                container_registry_tls_verify=True,
                containers_config_auth_file=None,
                container_command_short_name=None
            )
        ]
        container_registry_login_mock.assert_has_calls(calls)

    @patch('ploigos_step_runner.utils.containers.container_registry_login')
    def test_list_of_dicts_with_tls_verify_value(self, container_registry_login_mock):
        registries = [
            {
                'uri': 'registry.redhat.io',
                'username': 'hello1@world.xyz',
                'password': 'nope1',
                'tls-verify': False
            },
            {
                'uri': 'registry.internal.example.xyz',
                'username': 'hello2@example.xyz',
                'password': 'nope2',
                'tls-verify': True
            }
        ]

        container_registries_login(registries)

        calls = [
            call(
                container_registry_uri='registry.redhat.io',
                container_registry_username='hello1@world.xyz',
                container_registry_password='nope1',
                container_registry_tls_verify=False,
                containers_config_auth_file=None,
                container_command_short_name=None
            ),
            call(
                container_registry_uri='registry.internal.example.xyz',
                container_registry_username='hello2@example.xyz',
                container_registry_password='nope2',
                container_registry_tls_verify=True,
                containers_config_auth_file=None,
                container_command_short_name=None
            )
        ]
        container_registry_login_mock.assert_has_calls(calls)

    @patch('ploigos_step_runner.utils.containers.container_registry_login')
    def test_list_of_dicts_with_containers_config_auth_file(self, container_registry_login_mock):
        registries = [
            {
                'uri': 'registry.redhat.io',
                'username': 'hello1@world.xyz',
                'password': 'nope1'
            },
            {
                'uri': 'registry.internal.example.xyz',
                'username': 'hello2@example.xyz',
                'password': 'nope2'
            }
        ]

        container_registries_login(registries, '/tmp/mock/auth.json')

        calls = [
            call(
                container_registry_uri='registry.redhat.io',
                container_registry_username='hello1@world.xyz',
                container_registry_password='nope1',
                container_registry_tls_verify=True,
                containers_config_auth_file='/tmp/mock/auth.json',
                container_command_short_name=None
            ),
            call(
                container_registry_uri='registry.internal.example.xyz',
                container_registry_username='hello2@example.xyz',
                container_registry_password='nope2',
                container_registry_tls_verify=True,
                containers_config_auth_file='/tmp/mock/auth.json',
                container_command_short_name=None
            )
        ]
        container_registry_login_mock.assert_has_calls(calls)

    @patch('ploigos_step_runner.utils.containers.container_registry_login')
    def test_list_of_config_value(self, container_registry_login_mock):
        registries = [
            ConfigValue({
                'uri': 'registry.redhat.io',
                'username': 'hello1@world.xyz',
                'password': 'nope1'
            }),
            ConfigValue({
                'uri': 'registry.internal.example.xyz',
                'username': 'hello2@example.xyz',
                'password': 'nope2'
            })
        ]

        container_registries_login(registries)

        calls = [
            call(
                container_registry_uri='registry.redhat.io',
                container_registry_username='hello1@world.xyz',
                container_registry_password='nope1',
                container_registry_tls_verify=True,
                containers_config_auth_file=None,
                container_command_short_name=None
            ),
            call(
                container_registry_uri='registry.internal.example.xyz',
                container_registry_username='hello2@example.xyz',
                container_registry_password='nope2',
                container_registry_tls_verify=True,
                containers_config_auth_file=None,
                container_command_short_name=None
            )
        ]
        container_registry_login_mock.assert_has_calls(calls)

    def test_dict_of_dicts_missing_username(self):
        registries = {
            'registry.redhat.io': {
                'password': 'nope1'
            }
        }

        with self.assertRaisesRegex(
            AssertionError,
            r"Configuration for container registry \(registry.redhat.io\) "
            r"must specify a 'username': {'password': 'nope1'}"
        ):
            container_registries_login(registries)

    def test_dict_of_dicts_missing_password(self):
        registries = {
            'registry.redhat.io': {
                'username': 'hello1@world.xyz'
            }
        }

        with self.assertRaisesRegex(
            AssertionError,
            r"Configuration for container registry \(registry.redhat.io\) "
            r"must specify a 'password': {'username': 'hello1@world.xyz'}"
        ):
            container_registries_login(registries)

    def test_dict_of_dicts_missing_username(self):
        registries = {
            'registry.redhat.io': {
                'password': 'nope1'
            }
        }

        with self.assertRaisesRegex(
            AssertionError,
            r"Configuration for container registry \(registry.redhat.io\) "
            r"must specify a 'username': {'password': 'nope1'}"
        ):
            container_registries_login(registries)

    def test_list_of_dicts_missing_uri(self):
        registries = [
            {
                'username': 'hello1@world.xyz',
                'password': 'nope1'
            }
        ]

        with self.assertRaisesRegex(
            AssertionError,
            r"Configuration for container registry "
            r"must specify a 'uri': {'username': 'hello1@world.xyz', 'password': 'nope1'}"
        ):
            container_registries_login(registries)

    def test_list_of_dicts_missing_password(self):
        registries = [
            {
                'uri': 'registry.redhat.io',
                'username': 'hello1@world.xyz'
            }
        ]

        with self.assertRaisesRegex(
            AssertionError,
            r"Configuration for container registry "
            r"must specify a 'password': {'uri': 'registry.redhat.io', 'username': 'hello1@world.xyz'}"
        ):
            container_registries_login(registries)

    def test_list_of_dicts_missing_username(self):
        registries = [
            {
                'uri': 'registry.redhat.io',
                'password': 'nope1'
            }
        ]

        with self.assertRaisesRegex(
            AssertionError,
            r"Configuration for container registry "
            r"must specify a 'username': {'uri': 'registry.redhat.io', 'password': 'nope1'}"
        ):
            container_registries_login(registries)

    @patch('ploigos_step_runner.utils.containers.container_registry_login')
    def test_registries_none(self, container_registry_login):
        registries = None

        container_registries_login(registries)

        container_registry_login.assert_not_called()

    @patch('ploigos_step_runner.utils.containers.container_registry_login')
    def test_given_command_short_name_exists(self, container_registry_login_mock):
        registries = {
            'registry.redhat.io': {
                'username': 'hello1@world.xyz',
                'password': 'nope1'
            },
            'registry.internal.example.xyz': {
                'username': 'hello2@example.xyz',
                'password': 'nope2'
            }
        }

        container_registries_login(
            registries=registries,
            container_command_short_name='fake-podman'
        )

        calls = [
            call(
                container_registry_uri='registry.redhat.io',
                container_registry_username='hello1@world.xyz',
                container_registry_password='nope1',
                container_registry_tls_verify=True,
                containers_config_auth_file=None,
                container_command_short_name='fake-podman'
            ),
            call(
                container_registry_uri='registry.internal.example.xyz',
                container_registry_username='hello2@example.xyz',
                container_registry_password='nope2',
                container_registry_tls_verify=True,
                containers_config_auth_file=None,
                container_command_short_name='fake-podman'
            )
        ]
        container_registry_login_mock.assert_has_calls(calls)


class TestContainerRegistriesLoginForceOverrideTlsVerifyLogin(BaseTestCase):
    @patch('ploigos_step_runner.utils.containers.container_registry_login')
    def test_dict_of_dicts(self, container_registry_login_mock):
        registries = {
            'registry.redhat.io': {
                'username': 'hello1@world.xyz',
                'password': 'nope1'
            },
            'registry.internal.example.xyz': {
                'username': 'hello2@example.xyz',
                'password': 'nope2'
            }
        }

        container_registries_login(registries, containers_config_tls_verify=False)

        calls = [
            call(
                container_registry_uri='registry.redhat.io',
                container_registry_username='hello1@world.xyz',
                container_registry_password='nope1',
                container_registry_tls_verify=False,
                containers_config_auth_file=None,
                container_command_short_name=None
            ),
            call(
                container_registry_uri='registry.internal.example.xyz',
                container_registry_username='hello2@example.xyz',
                container_registry_password='nope2',
                container_registry_tls_verify=False,
                containers_config_auth_file=None,
                container_command_short_name=None
            )
        ]
        container_registry_login_mock.assert_has_calls(calls)

    @patch('ploigos_step_runner.utils.containers.container_registry_login')
    def test_list_of_dicts(self, container_registry_login_mock):
        registries = [
            {
                'uri': 'registry.redhat.io',
                'username': 'hello1@world.xyz',
                'password': 'nope1'
            },
            {
                'uri': 'registry.internal.example.xyz',
                'username': 'hello2@example.xyz',
                'password': 'nope2'
            }
        ]

        container_registries_login(registries, containers_config_tls_verify=False)

        calls = [
            call(
                container_registry_uri='registry.redhat.io',
                container_registry_username='hello1@world.xyz',
                container_registry_password='nope1',
                container_registry_tls_verify=False,
                containers_config_auth_file=None,
                container_command_short_name=None
            ),
            call(
                container_registry_uri='registry.internal.example.xyz',
                container_registry_username='hello2@example.xyz',
                container_registry_password='nope2',
                container_registry_tls_verify=False,
                containers_config_auth_file=None,
                container_command_short_name=None
            )
        ]
        container_registry_login_mock.assert_has_calls(calls)


    @patch('ploigos_step_runner.utils.containers.container_registry_login')
    def test_list_of_dicts_truthy(self, container_registry_login_mock):
        registries = [
            {
                'uri': 'registry.redhat.io',
                'username': 'hello1@world.xyz',
                'password': 'nope1'
            },
            {
                'uri': 'registry.internal.example.xyz',
                'username': 'hello2@example.xyz',
                'password': 'nope2'
            }
        ]

        container_registries_login(registries, containers_config_tls_verify=1)

        calls = [
            call(
                container_registry_uri='registry.redhat.io',
                container_registry_username='hello1@world.xyz',
                container_registry_password='nope1',
                container_registry_tls_verify=True,
                containers_config_auth_file=None,
                container_command_short_name=None
            ),
            call(
                container_registry_uri='registry.internal.example.xyz',
                container_registry_username='hello2@example.xyz',
                container_registry_password='nope2',
                container_registry_tls_verify=True,
                containers_config_auth_file=None,
                container_command_short_name=None
            )
        ]
        container_registry_login_mock.assert_has_calls(calls)

    @patch('ploigos_step_runner.utils.containers.container_registry_login')
    def test_list_of_dicts_falsey(self, container_registry_login_mock):
        registries = [
            {
                'uri': 'registry.redhat.io',
                'username': 'hello1@world.xyz',
                'password': 'nope1'
            },
            {
                'uri': 'registry.internal.example.xyz',
                'username': 'hello2@example.xyz',
                'password': 'nope2'
            }
        ]

        container_registries_login(registries, containers_config_tls_verify=0)

        calls = [
            call(
                container_registry_uri='registry.redhat.io',
                container_registry_username='hello1@world.xyz',
                container_registry_password='nope1',
                container_registry_tls_verify=False,
                containers_config_auth_file=None,
                container_command_short_name=None
            ),
            call(
                container_registry_uri='registry.internal.example.xyz',
                container_registry_username='hello2@example.xyz',
                container_registry_password='nope2',
                container_registry_tls_verify=False,
                containers_config_auth_file=None,
                container_command_short_name=None
            )
        ]
        container_registry_login_mock.assert_has_calls(calls)

class Test_create_container_from_image(BaseTestCase):
    @patch('sh.buildah', create=True)
    def test_success_default_repository_type(self, buildah_mock):
        # setup test
        image_tag = "localhost/mock-org/mock-image:v0.42.0-mock"

        # run test
        buildah_mock.side_effect = create_sh_side_effect(
            mock_stdout='mock-image-working-container-mock'
        )
        actual_container_name = create_container_from_image(
            image_tag=image_tag
        )

        # verify
        self.assertEqual(actual_container_name, 'mock-image-working-container-mock')

        buildah_mock.assert_called_once_with(
            'from',
            f"container-storage:{image_tag}",
            _out=Any(IOBase),
            _err=Any(IOBase),
            _tee='err'
        )

    @patch('sh.buildah', create=True)
    def test_success_remote_repository_type(self, buildah_mock):
        # setup test
        image_tag = "quay.io/mock-org/mock-image:v0.42.0-mock"

        # run test
        buildah_mock.side_effect = create_sh_side_effect(
            mock_stdout='mock-image-working-container-mock'
        )
        actual_container_name = create_container_from_image(
            image_tag=image_tag,
            repository_type='docker://'
        )

        # verify
        self.assertEqual(actual_container_name, 'mock-image-working-container-mock')

        buildah_mock.assert_called_once_with(
            'from',
            f"docker://{image_tag}",
            _out=Any(IOBase),
            _err=Any(IOBase),
            _tee='err'
        )

    @patch('sh.buildah', create=True)
    def test_error(self, buildah_mock):
        # setup test
        image_tag = "localhost/mock-org/mock-image:v0.42.0-mock"

        # run test with mock error
        with self.assertRaisesRegex(
            RuntimeError,
            re.compile(
                rf"Error creating container from image \({image_tag}\):"
                r".*RAN: buildah"
                r".*STDOUT:"
                r".*mock out"
                r".*STDERR:"
                r".*mock error",
                re.DOTALL
            )
        ):
            buildah_mock.side_effect = sh.ErrorReturnCode('buildah', b'mock out', b'mock error')
            create_container_from_image(
                image_tag=image_tag
            )

        # verify
        buildah_mock.assert_called_once_with(
            'from',
            f"container-storage:{image_tag}",
            _out=Any(IOBase),
            _err=Any(IOBase),
            _tee='err'
        )

class Test_mount_container(BaseTestCase):
    @patch('sh.buildah', create=True)
    def test_success(self, buildah_mock):
        buildah_unshare_command = sh.buildah.bake('unshare')
        container_name = "test"

        expected_mount_path = '/this/is/a/path'
        buildah_mock.bake('unshare').bake('buildah', 'mount').side_effect = create_sh_side_effect(
            mock_stdout=f"{expected_mount_path}",
        )

        container_mount_path = mount_container(
            buildah_unshare_command=buildah_unshare_command,
            container_id=container_name
        )

        self.assertEqual(container_mount_path, expected_mount_path)

        buildah_mock.bake('unshare').bake('buildah', 'mount').assert_called_once_with(
            container_name,
            _out=Any(IOBase),
            _err=Any(IOBase),
            _tee='err'
        )

    @patch('sh.buildah', create=True)
    def test_buildah_error(self, buildah_mock):
        buildah_unshare_command = sh.buildah.bake('unshare')
        container_name = "test"

        buildah_mock.bake('unshare').bake('buildah', 'mount').side_effect = sh.ErrorReturnCode(
            'buildah mount',
            b'mock out',
            b'mock error'
        )

        with self.assertRaisesRegex(
                RuntimeError,
                re.compile(
                    rf"Error mounting container \({container_name}\):"
                    r".*RAN: buildah"
                    r".*STDOUT:"
                    r".*mock out"
                    r".*STDERR:"
                    r".*mock error",
                    re.DOTALL
                )
        ):
            mount_container(
                buildah_unshare_command=buildah_unshare_command,
                container_id=container_name
            )

        buildah_mock.bake('unshare').bake('buildah', 'mount').assert_called_once_with(
            container_name,
            _out=Any(IOBase),
            _err=Any(IOBase),
            _tee='err'
        )

class Test_determine_container_image_build_tag_info(BaseTestCase):
    def test_given_image_version(self):
        actual_build_full_tag, actual_build_short_tag, actual_image_registry_uri, \
            actual_image_repository, actual_image_version = \
            determine_container_image_build_tag_info(
                image_version='1.0-123abc',
                organization='mock-org',
                application_name='mock-app',
                service_name='mock-service'
            )

        self.assertEqual(
            actual_build_full_tag,
            'localhost/mock-org/mock-app-mock-service:1.0-123abc'
        )
        self.assertEqual(
            actual_build_short_tag,
            'mock-org/mock-app-mock-service:1.0-123abc'
        )
        self.assertEqual(
            actual_image_registry_uri,
            'localhost'
        )
        self.assertEqual(
            actual_image_repository,
            'mock-app-mock-service'
        )
        self.assertEqual(
            actual_image_version,
            '1.0-123abc'
        )

    def test_default_image_version(self):
        actual_build_full_tag, actual_build_short_tag, actual_image_registry_uri, \
            actual_image_repository, actual_image_version = \
            determine_container_image_build_tag_info(
                image_version=None,
                organization='mock-org',
                application_name='mock-app',
                service_name='mock-service'
            )

        self.assertEqual(
            actual_build_full_tag,
            'localhost/mock-org/mock-app-mock-service:latest'
        )
        self.assertEqual(
            actual_build_short_tag,
            'mock-org/mock-app-mock-service:latest'
        )
        self.assertEqual(
            actual_image_registry_uri,
            'localhost'
        )
        self.assertEqual(
            actual_image_repository,
            'mock-app-mock-service'
        )
        self.assertEqual(
            actual_image_version,
            'latest'
        )



import os
from os import path
from unittest.mock import Mock, patch

from ploigos_step_runner.utils.xml import *
from testfixtures import TempDirectory
from tests.helpers.base_test_case import BaseTestCase


class TestXMLUtils_other(BaseTestCase):
    """Test for XML Utilities"""
    def test_get_xml_element_from_pom_file(self):
        """Test getting an xml element from the pom file."""
        with TempDirectory() as temp_dir:
            pom = b'''<project>
                        <modelVersion>4.0.0</modelVersion>
                        <groupId>com.mycompany.app</groupId>
                        <artifactId>my-app</artifactId>
                        <version>42.1</version>
                    </project>'''
            temp_dir.write('pom.xml', pom)
            pom_file_path = path.join(temp_dir.path, 'pom.xml')

            version = get_xml_element(pom_file_path, 'version').text
            artifact_id = get_xml_element(pom_file_path, 'artifactId').text
            group_id = get_xml_element(pom_file_path, 'groupId').text

            assert version == '42.1'
            assert artifact_id == 'my-app'
            assert group_id == 'com.mycompany.app'

    def test_get_xml_element_from_pom_file_with_namespace(self):
        """Test get xml element from pom file with namespace."""
        with TempDirectory() as temp_dir:
            pom = b'''<?xml version="1.0"?>
                      <project xsi:schemaLocation="http://maven.apache.org/POM/4.0.0 http://maven.apache.org/xsd/maven-4.0.0.xsd"
                          xmlns="http://maven.apache.org/POM/4.0.0"
                          xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
                          <modelVersion>4.0.0</modelVersion>
                          <groupId>com.mycompany.app</groupId>
                          <artifactId>my-app</artifactId>
                          <version>42.1</version>
                      </project>'''

            temp_dir.write('pom.xml', pom)
            pom_file_path = path.join(temp_dir.path, 'pom.xml')

            version = get_xml_element(pom_file_path, 'version').text
            artifact_id = get_xml_element(pom_file_path, 'artifactId').text
            group_id = get_xml_element(pom_file_path, 'groupId').text

            assert version == '42.1'
            assert artifact_id == 'my-app'
            assert group_id == 'com.mycompany.app'

    def test_get_xml_element_none_existent_file(self):
        """Test get xml element but file does not exist."""
        with self.assertRaisesRegex(
                ValueError,
                r"Given xml file does not exist: does_not_exist/pom.xml"):

            # pylint: disable=expression-not-assigned
            get_xml_element('does_not_exist/pom.xml', 'version').text

    def test_get_xml_element_element_does_not_exist(self):
        """Test get xml element but does not exist."""
        with TempDirectory() as temp_dir:
            pom = b'''<project>
                          <modelVersion>4.0.0</modelVersion>
                          <groupId>com.mycompany.app</groupId>
                          <artifactId>my-app</artifactId>
                          <version>42.1</version>
                      </project>'''

            temp_dir.write('pom.xml', pom)
            pom_file_path = path.join(temp_dir.path, 'pom.xml')

            with self.assertRaisesRegex(
                    ValueError,
                    r"Given xml file \(.*\) does not have ./does-not-exist element"):

                get_xml_element(pom_file_path, 'does-not-exist')

    def test_parse_xml_element_for_attributes(self):
        """Test to get an xml attribute from xml element

        """

        sample_file_path = os.path.join(
            os.path.dirname(__file__),
            'files',
            'sample.xml'
        )

        element = 'testsuite'
        attribs = ['time', 'tests']

        results = aggregate_xml_element_attribute_values(sample_file_path, element, attribs)

        expected_results = {'tests': 3, 'time': 4.485}

        self.assertEqual(results, expected_results)

    def test_parse_xml_element_for_attributes_multiple_files(self):
        """Test that element attributes are aggregated if multiple files with same element/attributes exist.

        """

        sample_file_path = os.path.join(
            os.path.dirname(__file__),
            'files/multiple_xml'
        )

        element = 'testsuite'
        attribs = ['time', 'tests','quality']

        results = aggregate_xml_element_attribute_values(sample_file_path, element, attribs)

        expected_results = {'time': 8.97, 'tests': 6, 'quality': ['42', 'medium-rare']}

        self.assertCountEqual(results, expected_results)

    def test_parse_xml_element_for_attributes_element_not_found(self):
        """Test that function returns empty dict if element is not found in file.
        """

        sample_file_path = os.path.join(
            os.path.dirname(__file__),
            'files',
            'sample.xml'
        )

        element = 'testsuite2'
        attribs = ['time', 'tests']

        results = aggregate_xml_element_attribute_values(sample_file_path, element, attribs)

        expected_results = {}

        self.assertEqual(results, expected_results)

    def test_parse_xml_element_for_attributes_some_there_some_missing_attributes(self):
        """Test that function returns empty dict if attribute is not found in element in file.

        Should get the values that are there and ignore the ones that aren't.
        """

        sample_file_path = os.path.join(
            os.path.dirname(__file__),
            'files',
            'sample.xml'
        )

        element = 'testsuite'
        attribs = ['time', 'tests', 'mock1']

        results = aggregate_xml_element_attribute_values(sample_file_path, element, attribs)

        expected_results = {'time': 4.485, 'tests': 3}

        self.assertEqual(results, expected_results)

    def test_parse_xml_element_for_attributes_all_missing_attributes(self):
        """Test that function returns empty dict if attribute is not found in element in file.
        """

        sample_file_path = os.path.join(
            os.path.dirname(__file__),
            'files',
            'sample.xml'
        )

        element = 'testsuite'
        attribs = ['mock1', 'mock2']

        results = aggregate_xml_element_attribute_values(sample_file_path, element, attribs)

        expected_results = {}

        self.assertEqual(results, expected_results)

class TestXMLUtils_get_xml_element_by_path(BaseTestCase):
    def test_none_existent_file(self):
        """Test get xml element by xpath but file does not exist."""
        with self.assertRaisesRegex(
            ValueError,
            r"Given xml file does not exist: does_not_exist/pom.xml"
        ):
            get_xml_element_by_path('does_not_exist/pom.xml', 'build').text

    def test__exists(self):
        """Test to get an xml element where it exists."""
        with TempDirectory() as temp_dir:
            pom = b'''<project
                        xsi:schemaLocation="http://maven.apache.org/POM/4.0.0 http://maven.apache.org/xsd/maven-4.0.0.xsd"
                        xmlns="http://maven.apache.org/POM/4.0.0"
                        xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
                        <modelVersion>4.0.0</modelVersion>
                        <groupId>com.mycompany.app</groupId>
                        <artifactId>my-app</artifactId>
                        <version>42.1</version>
                        <build>
                            <plugins>
                                <plugin>
                                    <artifactId>maven-surefire-plugin</artifactId>
                                    <configuration>
                                        <reportsDirectory>mycompany-reports-directory</reportsDirectory>
                                    </configuration>
                                </plugin>
                            </plugins>
                        </build>
                    </project>'''

            temp_dir.write('pom.xml', pom)
            pom_file_path = path.join(temp_dir.path, 'pom.xml')
            element = get_xml_element_by_path(
                pom_file_path,
                'mvn:build/mvn:plugins/mvn:plugin/[mvn:artifactId="maven-surefire-plugin"]/'\
                    'mvn:configuration/mvn:reportsDirectory',
                default_namespace='mvn'
            )

            self.assertEqual(element.text, 'mycompany-reports-directory')

    def test_exists_no_namespaces(self):
        """Test to get an xml element where it exists, but no namespaces exist."""
        with TempDirectory() as temp_dir:
            pom = b'''<project>
                        <modelVersion>4.0.0</modelVersion>
                        <groupId>com.mycompany.app</groupId>
                        <artifactId>my-app</artifactId>
                        <version>42.1</version>
                        <build>
                            <plugins>
                                <plugin>
                                    <artifactId>maven-surefire-plugin</artifactId>
                                    <configuration>
                                        <reportsDirectory>mycompany-reports-directory</reportsDirectory>
                                    </configuration>
                                </plugin>
                            </plugins>
                        </build>
                    </project>'''

            temp_dir.write('pom.xml', pom)
            pom_file_path = path.join(temp_dir.path, 'pom.xml')
            element = get_xml_element_by_path(
                pom_file_path,
                'build/plugins/plugin/[artifactId="maven-surefire-plugin"]/' \
                    'configuration/reportsDirectory'
            )

            self.assertEqual(element.text, 'mycompany-reports-directory')

    def test_element_not_found(self):
        """Test to get an xml element where it does not exist."""
        with TempDirectory() as temp_dir:
            pom = b'''<project
                        xsi:schemaLocation="http://maven.apache.org/POM/4.0.0 http://maven.apache.org/xsd/maven-4.0.0.xsd"
                        xmlns="http://maven.apache.org/POM/4.0.0"
                        xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
                        <modelVersion>4.0.0</modelVersion>
                        <groupId>com.mycompany.app</groupId>
                        <artifactId>my-app</artifactId>
                        <version>42.1</version>
                        <build>
                            <plugins>
                                <plugin>
                                    <artifactId>maven-surefire-plugin</artifactId>
                                    <configuration>
                                        <reportsDirectory>mycompany-reports-directory</reportsDirectory>
                                    </configuration>
                                </plugin>
                            </plugins>
                        </build>
                    </project>'''

            temp_dir.write('pom.xml', pom)
            pom_file_path = path.join(temp_dir.path, 'pom.xml')
            element = get_xml_element_by_path(
                pom_file_path,
                'mvn:this_does_not_exist',
                default_namespace='mvn'
            )

            self.assertIsNone(element)

    def test_element_not_found_no_namespaces(self):
        """Test to get an xml element where it does not exist and no namespaces exist."""
        with TempDirectory() as temp_dir:
            pom = b'''<project>
                        <modelVersion>4.0.0</modelVersion>
                        <groupId>com.mycompany.app</groupId>
                        <artifactId>my-app</artifactId>
                        <version>42.1</version>
                        <build>
                            <plugins>
                                <plugin>
                                    <artifactId>maven-surefire-plugin</artifactId>
                                    <configuration>
                                        <reportsDirectory>mycompany-reports-directory</reportsDirectory>
                                    </configuration>
                                </plugin>
                            </plugins>
                        </build>
                    </project>'''

            temp_dir.write('pom.xml', pom)
            pom_file_path = path.join(temp_dir.path, 'pom.xml')
            element = get_xml_element_by_path(
                pom_file_path,
                'this-does-not-exist'
            )

            self.assertIsNone(element)

    def test_find_all_single_result(self):
        with TempDirectory() as temp_dir:
            pom = b'''<project>
    <modelVersion>4.0.0</modelVersion>
    <groupId>com.mycompany.app</groupId>
    <artifactId>my-app</artifactId>
    <version>42.1</version>
    <build>
        <plugins>
            <plugin>
                <groupId>org.apache.maven.plugins</groupId>
                <artifactId>maven-surefire-plugin</artifactId>
                <version>${surefire-plugin.version}</version>
                <configuration>
                    <skipTests>true</skipTests>
                </configuration>
                <executions>
                    <execution>
                        <id>unit-tests</id>
                        <phase>test</phase>
                        <goals>
                            <goal>test</goal>
                        </goals>
                        <configuration>
                            <skipTests>${skipTests}</skipTests>
                            <reportsDirectory>${project.build.directory}/surefire-reports-unit-test</reportsDirectory>
                        </configuration>
                    </execution>
                    <execution>
                        <id>unit-tests2</id>
                        <phase>test</phase>
                        <goals>
                            <goal>test</goal>
                        </goals>
                        <configuration>
                            <skipTests>${skipTests}</skipTests>
                            <reportsDirectory>${project.build.directory}/surefire-reports-unit-test2</reportsDirectory>
                        </configuration>
                    </execution>
                    <execution>
                        <id>integration-tests</id>
                        <phase>integration-test</phase>
                        <goals>
                            <goal>test</goal>
                        </goals>
                        <configuration>
                            <skipTests>${skipITs}</skipTests>
                            <reportsDirectory>${project.build.directory}/surefire-reports-uat</reportsDirectory>
                            <systemProperties>
                            <java.util.logging.manager>org.jboss.logmanager.LogManager</java.util.logging.manager>
                            </systemProperties>
                            <includes>
                            <include>**/*IT.*</include>
                            </includes>
                        </configuration>
                    </execution>
                </executions>
            </plugin>
        </plugins>
    </build>
</project>'''

            temp_dir.write('pom.xml', pom)
            pom_file_path = path.join(temp_dir.path, 'pom.xml')
            elements = get_xml_element_by_path(
                pom_file_path,
                'build/plugins/plugin/[artifactId="maven-surefire-plugin"]/executions/execution/[phase="integration-test"]/configuration/reportsDirectory',
                find_all=True
            )

            self.assertEqual(len(elements), 1)
            self.assertEqual(
                elements[0].text,
                '${project.build.directory}/surefire-reports-uat'
            )

    def test_find_all_multiple_results(self):
        with TempDirectory() as temp_dir:
            pom = b'''<project>
    <modelVersion>4.0.0</modelVersion>
    <groupId>com.mycompany.app</groupId>
    <artifactId>my-app</artifactId>
    <version>42.1</version>
    <build>
        <plugins>
            <plugin>
                <groupId>org.apache.maven.plugins</groupId>
                <artifactId>maven-surefire-plugin</artifactId>
                <version>${surefire-plugin.version}</version>
                <configuration>
                    <skipTests>true</skipTests>
                </configuration>
                <executions>
                    <execution>
                        <id>unit-tests</id>
                        <phase>test</phase>
                        <goals>
                            <goal>test</goal>
                        </goals>
                        <configuration>
                            <skipTests>${skipTests}</skipTests>
                            <reportsDirectory>${project.build.directory}/surefire-reports-unit-test</reportsDirectory>
                        </configuration>
                    </execution>
                    <execution>
                        <id>unit-tests2</id>
                        <phase>test</phase>
                        <goals>
                            <goal>test</goal>
                        </goals>
                        <configuration>
                            <skipTests>${skipTests}</skipTests>
                            <reportsDirectory>${project.build.directory}/surefire-reports-unit-test2</reportsDirectory>
                        </configuration>
                    </execution>
                    <execution>
                        <id>integration-tests</id>
                        <phase>integration-test</phase>
                        <goals>
                            <goal>test</goal>
                        </goals>
                        <configuration>
                            <skipTests>${skipITs}</skipTests>
                            <reportsDirectory>${project.build.directory}/surefire-reports-uat</reportsDirectory>
                            <systemProperties>
                            <java.util.logging.manager>org.jboss.logmanager.LogManager</java.util.logging.manager>
                            </systemProperties>
                            <includes>
                            <include>**/*IT.*</include>
                            </includes>
                        </configuration>
                    </execution>
                </executions>
            </plugin>
        </plugins>
    </build>
</project>'''

            temp_dir.write('pom.xml', pom)
            pom_file_path = path.join(temp_dir.path, 'pom.xml')
            elements = get_xml_element_by_path(
                pom_file_path,
                'build/plugins/plugin/[artifactId="maven-surefire-plugin"]/executions/' \
                    'execution/[phase="test"]/configuration/reportsDirectory',
                find_all=True
            )

            self.assertEqual(len(elements), 2)
            self.assertEqual(
                elements[0].text,
                '${project.build.directory}/surefire-reports-unit-test'
            )
            self.assertEqual(
                elements[1].text,
                '${project.build.directory}/surefire-reports-unit-test2'
            )

@patch('ploigos_step_runner.utils.xml.get_xml_element_by_path')
class TestXMLUtils_get_xml_element_text_by_path(BaseTestCase):
    def test_find_one(self, mock_get_xml_element_by_path):
        # set up mock
        mock_get_xml_element_by_path.return_value = Mock(text='mock element text')

        # run test
        actual_results = get_xml_element_text_by_path(
            xml_file_path='/mock.xml',
            xpath='mock-path',
            default_namespace=None,
            xml_namespace_dict=None,
            find_all=False
        )

        # do verification
        self.assertEqual(actual_results, 'mock element text')
        mock_get_xml_element_by_path.assert_called_once_with(
            xml_file_path='/mock.xml',
            xpath='mock-path',
            default_namespace=None,
            xml_namespace_dict=None,
            find_all=False
        )

    def test_find_multiple(self, mock_get_xml_element_by_path):
        # set up mock
        mock_get_xml_element_by_path.return_value = [
            Mock(text='mock element text 1'),
            Mock(text='mock element text 2')
        ]

        # run test
        actual_results = get_xml_element_text_by_path(
            xml_file_path='/mock.xml',
            xpath='mock-path',
            default_namespace=None,
            xml_namespace_dict=None,
            find_all=True
        )

        # do verification
        self.assertEqual(actual_results, ['mock element text 1', 'mock element text 2'])

        mock_get_xml_element_by_path.assert_called_once_with(
            xml_file_path='/mock.xml',
            xpath='mock-path',
            default_namespace=None,
            xml_namespace_dict=None,
            find_all=True
        )



import pytest

from ploigos_step_runner import StepRunnerException

def raise_StepRunnerException():
    raise StepRunnerException('test')

def test_StepRunnerException():
    with pytest.raises(StepRunnerException):
        raise_StepRunnerException()
